"""Tests for the connection module."""

import asyncio
import json
from typing import Any, Optional, Tuple
from unittest.mock import AsyncMock, Mock, patch

import aiohttp

import pytest
from asusrouter.connection import Connection, generate_credentials
from asusrouter.const import DEFAULT_TIMEOUT, USER_AGENT, RequestType
from asusrouter.error import (
    AsusRouter404Error,
    AsusRouterAccessError,
    AsusRouterConnectionError,
    AsusRouterError,
    AsusRouterLogoutError,
    AsusRouterTimeoutError,
)
from asusrouter.modules.endpoint import EndpointService


@pytest.mark.asyncio
async def test_generate_credentials() -> None:
    """Test the `generate_credentials`."""

    username = "user"
    password = "pass"
    payload, headers = generate_credentials(username, password)
    assert payload == "login_authorization=dXNlcjpwYXNz"
    assert headers == {"user-agent": "asusrouter--DUTUtil-"}


class TestConnection:
    def _assert_connection(
        self,
        conn: Connection,
        hostname: str,
        username: str,
        password: str,
        port: Optional[int],
        use_ssl: bool,
        session: Optional[Mock],
        timeout: int,
        dumpback: Optional[Mock],
        mock_new_session: Mock,
    ) -> None:
        """Test the Connection object."""

        assert conn._hostname == hostname
        assert conn._username == username
        assert conn._password == password
        assert conn._port == (port or (8443 if use_ssl else 80))
        assert conn._ssl == use_ssl
        assert conn._session == (session or mock_new_session.return_value)
        assert conn._timeout == timeout
        assert conn._dumpback == dumpback
        assert conn._token is None
        assert conn._header is None
        assert conn._connected is False
        assert isinstance(conn._connection_lock, asyncio.Lock)

    @pytest.mark.parametrize(
        "hostname, username, password, port, use_ssl, session, timeout,"
        "dumpback",
        [
            ("localhost", "user", "pass", None, False, None, None, None),
            ("localhost", "user", "pass", 8080, True, None, None, None),
            ("localhost", "user", "pass", 8080, False, Mock(), None, None),
            ("localhost", "user", "pass", 8080, True, Mock(), 10, Mock()),
            ("", "user", "pass", 8080, True, Mock(), None, Mock()),
            ("localhost", "", "pass", 8080, True, None, 10, None),
            ("localhost", "user", "pass", -1, True, None, 15, None),
        ],
    )
    @patch.object(Connection, "_new_session", return_value=Mock())
    def test_init(
        self,
        mock_new_session: Mock,
        hostname: str,
        username: str,
        password: str,
        port: int,
        use_ssl: bool,
        session: Optional[Mock],
        timeout: Optional[int],
        dumpback: Optional[Mock],
    ) -> None:
        """Test the initialization of the connection object."""

        # Create a Connection
        conn = Connection(
            hostname=hostname,
            username=username,
            password=password,
            port=port,
            use_ssl=use_ssl,
            session=session,
            timeout=timeout,
            dumpback=dumpback,
        )

        # Test the Connection
        self._assert_connection(
            conn,
            hostname=hostname,
            username=username,
            password=password,
            port=port,
            use_ssl=use_ssl,
            session=session,
            timeout=timeout or DEFAULT_TIMEOUT,
            dumpback=dumpback,
            mock_new_session=mock_new_session,
        )

    @patch.object(Connection, "_new_session", return_value=Mock())
    def test_init_multiple_instances(self, mock_new_session: Mock) -> None:
        """Test the initialization of multiple connection objects."""

        # Create two Connection
        conn1 = Connection(
            hostname="localhost1",
            username="user1",
            password="pass1",
            port=8080,
            use_ssl=True,
            timeout=None,
        )
        conn2 = Connection(
            hostname="localhost2",
            username="user2",
            password="pass2",
            port=9090,
            use_ssl=False,
            timeout=1,
        )

        # Test the Connections
        self._assert_connection(
            conn1,
            hostname="localhost1",
            username="user1",
            password="pass1",
            port=8080,
            use_ssl=True,
            session=None,
            timeout=DEFAULT_TIMEOUT,
            dumpback=None,
            mock_new_session=mock_new_session,
        )
        self._assert_connection(
            conn2,
            hostname="localhost2",
            username="user2",
            password="pass2",
            port=9090,
            use_ssl=False,
            session=None,
            timeout=1,
            dumpback=None,
            mock_new_session=mock_new_session,
        )

        # Ensure that instances do not interfere with each other
        assert conn1._hostname != conn2._hostname
        assert conn1._username != conn2._username
        assert conn1._password != conn2._password
        assert conn1._port != conn2._port
        assert conn1._ssl != conn2._ssl
        assert conn1._timeout != conn2._timeout

    @pytest.mark.asyncio
    async def test_aenter(self) -> None:
        """Test the `__aenter__` method."""

        with patch.object(
            Connection, "async_connect", new_callable=AsyncMock
        ) as mock_connect:
            # Create a Connection
            connection = Connection(
                hostname="localhost",
                username="user",
                password="pass",
                port=8080,
                use_ssl=True,
            )

            # Enter the context manager
            async with connection as conn:
                # Verify that async_connect was called
                mock_connect.assert_called_once()

                # Verify that the connection object is returned
                assert conn is connection

    @pytest.mark.asyncio
    async def test_aexit(self) -> None:
        """Test the `__aexit__` method."""

        # Mock the async_connect and async_close methods
        with (
            patch.object(
                Connection, "async_connect", new_callable=AsyncMock
            ) as mock_connect,
            patch.object(
                Connection, "async_close", new_callable=AsyncMock
            ) as mock_close,
        ):
            # Create a Connection
            connection = Connection(
                hostname="localhost",
                username="user",
                password="pass",
                port=8080,
                use_ssl=True,
            )

            # Exit the context manager
            async with connection:
                pass

            # Verify that async_connect was called
            mock_connect.assert_called_once()

            # Verify that async_close was called
            mock_close.assert_called_once()

    @pytest.mark.asyncio
    async def test_create(self) -> None:
        """Test the `create` class method."""

        # Mock the async_connect method
        with patch.object(
            Connection, "async_connect", new_callable=AsyncMock
        ) as mock_connect:
            # Define test parameters
            hostname = "localhost"
            username = "user"
            password = "pass"
            port = 8080
            use_ssl = True
            session = Mock()
            timeout = 15
            dumpback = Mock()

            # Call the create method
            connection = await Connection.create(
                hostname=hostname,
                username=username,
                password=password,
                port=port,
                use_ssl=use_ssl,
                session=session,
                timeout=timeout,
                dumpback=dumpback,
            )

            # Verify that async_connect was called
            mock_connect.assert_called_once()

            # Verify that the returned object is an instance of Connection
            assert isinstance(connection, Connection)

            # Verify that the connection object is initialized correctly
            assert connection._hostname == hostname
            assert connection._username == username
            assert connection._password == password
            assert connection._port == port
            assert connection._ssl == use_ssl
            assert connection._session == session
            assert connection._timeout == timeout
            assert connection._dumpback == dumpback
            assert connection._token is None
            assert connection._header is None
            assert connection._connected is False
            assert isinstance(connection._connection_lock, asyncio.Lock)

    @pytest.mark.asyncio
    async def test_new_session(self) -> None:
        """Test the `_new_session` method."""

        # Create a Connection
        connection = Connection(
            hostname="localhost",
            username="user",
            password="pass",
            timeout=30,
        )

        # Mock aiohttp.ClientSession and aiohttp.TCPConnector
        with (
            patch("aiohttp.ClientSession") as mock_client_session,
            patch("aiohttp.TCPConnector") as mock_tcp_connector,
            patch(
                "asusrouter.connection.get_cookie_jar",
            ) as mock_get_cookie_jar,
        ):
            # Use a Mock for the cookie jar
            mock_cookie_jar = Mock()
            mock_get_cookie_jar.return_value = mock_cookie_jar

            # Call _new_session
            session = connection._new_session()

            # Verify that _manage_session is set to True
            assert connection._manage_session is True

            # Verify that aiohttp.TCPConnector was called
            # with the correct SSL setting
            mock_tcp_connector.assert_called_once_with(
                ssl=connection._verify_ssl
            )

            # Verify that aiohttp.ClientSession was called
            # with the correct timeout and cookie jar
            mock_client_session.assert_called_once_with(
                connector=mock_tcp_connector.return_value,
                cookie_jar=mock_cookie_jar,
                timeout=aiohttp.ClientTimeout(total=connection._timeout),
            )

            # Verify that the returned session is the mocked session
            assert session == mock_client_session.return_value

    @pytest.mark.asyncio
    async def test_async_close(self) -> None:
        """Test the `async_close` method."""

        # Create a Connection
        connection = Connection(
            hostname="localhost",
            username="user",
            password="pass",
        )

        # Mock the session
        mock_session = AsyncMock()
        connection._session = mock_session
        connection._manage_session = True

        # Case 1: Session is not closed
        mock_session.closed = False
        await connection.async_close()
        mock_session.close.assert_awaited_once()

        # Case 2: Session is already closed
        mock_session.reset_mock()
        mock_session.closed = True
        await connection.async_close()
        mock_session.close.assert_not_awaited()

        # Case 3: _manage_session is False
        connection._manage_session = False
        mock_session.reset_mock()
        await connection.async_close()
        mock_session.close.assert_not_awaited()

    @pytest.mark.asyncio
    async def test_async_connect(self) -> None:
        """Test the `async_connect` method."""

        # Create a Connection
        connection = Connection(
            hostname="localhost",
            username="user",
            password="pass",
            timeout=30,
        )

        # Mock _async_connect_with_lock
        with patch.object(
            connection, "_async_connect_with_lock", new_callable=AsyncMock
        ) as mock_connect_with_lock:
            # Case 1: Connection succeeds
            mock_connect_with_lock.return_value = True
            result = await connection.async_connect()
            mock_connect_with_lock.assert_awaited_once_with(None)
            assert result is True

            # Case 2: Connection times out
            mock_connect_with_lock.reset_mock()
            mock_connect_with_lock.side_effect = asyncio.TimeoutError
            result = await connection.async_connect()
            mock_connect_with_lock.assert_awaited_once_with(None)
            assert result is False

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "connected, send_request_result, send_request_side_effect,"
        "expected_result, expected_exception, exception_message",
        [
            # Case 1: Already connected
            (True, None, None, True, None, None),
            # Case 2: Successful connection
            (
                False,
                (200, {}, '{"asus_token": "test_token"}'),
                None,
                True,
                None,
                None,
            ),
            # Case 3: AsusRouterAccessError occurs
            (
                False,
                None,
                AsusRouterAccessError("Access denied"),
                None,
                AsusRouterAccessError,
                "Cannot access EndpointService.LOGIN. "
                "Failed in `async_connect`",
            ),
            # Case 4: Unexpected error occurs
            (
                False,
                None,
                Exception("Unexpected error"),
                None,
                Exception,
                "Unexpected error",
            ),
            # Case 5: AsusRouterError occurs
            (
                False,
                None,
                AsusRouterError("Generic AsusRouter error"),
                None,
                AsusRouterError,
                "Generic AsusRouter error",
            ),
            # Case 6: No token received
            (
                False,
                (200, {}, '{"invalid_key": "no_token"}'),
                None,
                False,
                None,
                None,
            ),
        ],
    )
    async def test_async_connect_with_lock(
        self,
        connected: bool,
        send_request_result: Optional[tuple[int, Any, str]],
        send_request_side_effect: Optional[Exception],
        expected_result: Optional[bool],
        expected_exception: Optional[type[BaseException]],
        exception_message: Optional[str],
    ) -> None:
        """Test `_async_connect_with_lock` with parameterized scenarios."""

        # Create a Connection
        connection = Connection(
            hostname="localhost",
            username="user",
            password="pass",
        )

        # Set the initial connection state
        connection._connected = connected

        # Mock the `_send_request` method
        with patch.object(
            connection, "_send_request", new_callable=AsyncMock
        ) as mock_send_request:
            # Configure the mock based on the test case
            if send_request_side_effect:
                mock_send_request.side_effect = send_request_side_effect
            else:
                mock_send_request.return_value = send_request_result

            if expected_exception:
                # Verify that the expected exception is raised
                with pytest.raises(
                    expected_exception, match=exception_message
                ):
                    await connection._async_connect_with_lock()
            else:
                # Call `_async_connect_with_lock`
                result = await connection._async_connect_with_lock()

                # Verify the result
                assert result == expected_result

                # Verify that `_send_request` was called
                # if not already connected
                if not connected:
                    mock_send_request.assert_called_once_with(
                        EndpointService.LOGIN,
                        "login_authorization=dXNlcjpwYXNz",
                        {"user-agent": USER_AGENT},
                    )
                else:
                    mock_send_request.assert_not_called()

            # Additional assertions for successful connection
            if (
                not connected
                and not expected_exception
                and send_request_result
            ):
                if "asus_token" in json.loads(send_request_result[2]):
                    assert connection._connected is True
                    assert connection._token == "test_token"
                    assert connection._header == {
                        "user-agent": USER_AGENT,
                        "cookie": "asus_token=test_token",
                    }
                else:
                    assert connection._connected is False

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "connected, send_request_side_effect, expected_result,"
        "reset_connection_called",
        [
            # Case 1: Logout error is the correct result
            (True, AsusRouterLogoutError("Session is logged out"), True, True),
            # Case 2: Already disconnected
            (False, None, True, False),
            # Case X: Random results
            (True, None, False, False),
            (True, AsusRouterError("Generic error"), False, False),
        ],
    )
    async def test_async_disconnect(
        self,
        connected: bool,
        send_request_side_effect: Optional[Exception],
        expected_result: bool,
        reset_connection_called: bool,
    ) -> None:
        """Test the `async_disconnect` method with parameterized scenarios."""

        # Create a Connection
        connection = Connection(
            hostname="localhost",
            username="user",
            password="pass",
        )

        # Set the initial connection state
        connection._connected = connected

        # Mock the `_send_request` and `reset_connection` methods
        with (
            patch.object(
                connection, "_send_request", new_callable=AsyncMock
            ) as mock_send_request,
            patch.object(
                connection, "reset_connection", new_callable=Mock
            ) as mock_reset_connection,
        ):
            # Configure the mock for `_send_request`
            if send_request_side_effect:
                mock_send_request.side_effect = send_request_side_effect
            else:
                mock_send_request.return_value = None

            # Call `async_disconnect`
            result = await connection.async_disconnect()

            # Verify the result
            assert result == expected_result

            # Verify `_send_request` was called only if connected
            if connected:
                mock_send_request.assert_called_once_with(
                    EndpointService.LOGOUT
                )
            else:
                mock_send_request.assert_not_called()

            # Verify `reset_connection` was called if expected
            if reset_connection_called:
                mock_reset_connection.assert_called_once()
            else:
                mock_reset_connection.assert_not_called()

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "resp_status, resp_headers, resp_content, make_request_side_effect,"
        "expected_exception, exception_message",
        [
            # Case 1: Successful request
            (200, {"header": "value"}, '{"key": "value"}', None, None, None),
            # Case 2: 404 error
            (
                404,
                {},
                "",
                None,
                AsusRouter404Error,
                "Endpoint EndpointService.LOGIN not found",
            ),
            # Case 3: Non-200 status code
            (
                403,
                {},
                "",
                None,
                AsusRouterAccessError,
                "Cannot access EndpointService.LOGIN, status 403",
            ),
            # Case 4: Access error in content
            (
                200,
                {},
                '{"error_status": "8"}',
                None,
                AsusRouterAccessError,
                "Access error",
            ),
            # Case 5: Connection error (ClientConnectorError)
            (
                None,
                None,
                None,
                aiohttp.ClientConnectorError(Mock(), Mock()),
                AsusRouterConnectionError,
                "Cannot connect to `localhost` on port `80`.",
            ),
            # Case 6: Connection error (ClientConnectionError)
            (
                None,
                None,
                None,
                aiohttp.ClientConnectionError("Connection error"),
                AsusRouterConnectionError,
                r"Connection error",
            ),
            # Case 7: Connection error (ClientOSError)
            (
                None,
                None,
                None,
                aiohttp.ClientOSError("OS error"),
                AsusRouterConnectionError,
                r"OS error",
            ),
            # Case 8: Unexpected error
            (
                None,
                None,
                None,
                Exception("Unexpected error"),
                Exception,
                "Unexpected error",
            ),
            # Case 9: asyncio.TimeoutError
            (
                None,
                None,
                None,
                asyncio.TimeoutError("operation timed out"),
                AsusRouterTimeoutError,
                r"Data cannot be retrieved due to an asyncio error\. "
                r"Connection failed: operation timed out",
            ),
            # Case 10: asyncio.CancelledError
            (
                None,
                None,
                None,
                asyncio.CancelledError("operation cancelled"),
                AsusRouterTimeoutError,
                r"Data cannot be retrieved due to an asyncio error\. "
                r"Connection failed: operation cancelled",
            ),
        ],
    )
    async def test_send_request(
        self,
        resp_status: Optional[int],
        resp_headers: Optional[dict[str, str]],
        resp_content: Optional[str],
        make_request_side_effect: Optional[Exception],
        expected_exception: Optional[type[BaseException]],
        exception_message: Optional[str],
    ) -> None:
        """Test the `_send_request` method with parameterized scenarios."""

        # Create a Connection
        connection = Connection(
            hostname="localhost",
            username="user",
            password="pass",
        )

        # Mock the _make_request method
        with (
            patch.object(
                connection, "_make_request", new_callable=AsyncMock
            ) as mock_make_request,
            patch(
                "asusrouter.connection.handle_access_error"
            ) as mock_handle_access_error,
            patch.object(
                connection, "reset_connection", new_callable=Mock
            ) as mock_reset_connection,
        ):
            # Configure the mock for _make_request
            if make_request_side_effect:
                mock_make_request.side_effect = make_request_side_effect
            else:
                mock_make_request.return_value = (
                    resp_status,
                    resp_headers,
                    resp_content,
                )

            # Configure handle_access_error to raise an exception if called
            if resp_content and "error_status" in resp_content:
                mock_handle_access_error.side_effect = AsusRouterAccessError(
                    "Access error"
                )

            if expected_exception:
                with pytest.raises(
                    expected_exception, match=exception_message
                ):
                    await connection._send_request(EndpointService.LOGIN)
            else:
                result = await connection._send_request(EndpointService.LOGIN)
                assert result == (resp_status, resp_headers, resp_content)

            # Verify _make_request was called with expected parameters
            mock_make_request.assert_called_once_with(
                EndpointService.LOGIN, None, None, RequestType.POST
            )

            # Verify handle_access_error was called if error_status
            # in content, else not
            if resp_content and "error_status" in resp_content:
                mock_handle_access_error.assert_called_once_with(
                    EndpointService.LOGIN,
                    resp_status,
                    resp_headers,
                    resp_content,
                )
            else:
                mock_handle_access_error.assert_not_called()

            # For ClientConnectorError, reset_connection is expected,
            # but for asyncio.TimeoutError or CancelledError we do not
            # call reset_connection.
            if make_request_side_effect and isinstance(
                make_request_side_effect, aiohttp.ClientConnectorError
            ):
                mock_reset_connection.assert_called_once()
            else:
                mock_reset_connection.assert_not_called()

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "connected, connect_result, send_request_result,"
        "send_request_side_effect, expected_exception, exception_message",
        [
            # Case 1: Already connected, successful request
            (
                True,
                None,
                (200, {"header": "value"}, '{"key": "value"}'),
                None,
                None,
                None,
            ),
            # Case 2: Not connected, successful connection, successful request
            (
                False,
                True,
                (200, {"header": "value"}, '{"key": "value"}'),
                None,
                None,
                None,
            ),
            # Case 3: Not connected, connection fails
            (
                False,
                False,
                None,
                None,
                AsusRouterTimeoutError,
                "Data cannot be retrieved. Connection failed",
            ),
            # Case 4: Already connected, request fails
            (
                True,
                None,
                None,
                AsusRouterAccessError("Access denied"),
                AsusRouterAccessError,
                "Access denied",
            ),
        ],
    )
    async def test_async_query(
        self,
        connected: bool,
        connect_result: Optional[bool],
        send_request_result: Optional[Tuple[int, dict[str, str], str]],
        send_request_side_effect: Optional[Exception],
        expected_exception: Optional[type[BaseException]],
        exception_message: Optional[str],
    ) -> None:
        """Test the `async_query` method with parameterized scenarios."""

        # Create a Connection
        connection = Connection(
            hostname="localhost",
            username="user",
            password="pass",
        )

        # Set the initial connection state
        connection._connected = connected

        # Mock the `async_connect` and `_send_request` methods
        with (
            patch.object(
                connection, "async_connect", new_callable=AsyncMock
            ) as mock_async_connect,
            patch.object(
                connection, "_send_request", new_callable=AsyncMock
            ) as mock_send_request,
        ):
            # Configure the mock for `async_connect`
            if connect_result is not None:
                mock_async_connect.return_value = connect_result

            # Configure the mock for `_send_request`
            if send_request_side_effect:
                mock_send_request.side_effect = send_request_side_effect
            else:
                mock_send_request.return_value = send_request_result

            if not connected and connect_result:
                # Simulate successful connection by setting `_connected`
                # to True
                mock_async_connect.side_effect = lambda: setattr(
                    connection, "_connected", True
                )

            if expected_exception:
                # Verify that the expected exception is raised
                with pytest.raises(
                    expected_exception, match=exception_message
                ):
                    await connection.async_query(EndpointService.LOGIN)
            else:
                # Call `async_query`
                result = await connection.async_query(EndpointService.LOGIN)

                # Verify the result
                assert result == send_request_result

            # Verify `async_connect` was called if not connected
            if not connected:
                mock_async_connect.assert_called_once()
            else:
                mock_async_connect.assert_not_called()

            # Verify `_send_request` was called
            if connected or (not connected and connect_result):
                mock_send_request.assert_called_once_with(
                    EndpointService.LOGIN, None, None, RequestType.POST
                )
            else:
                mock_send_request.assert_not_called()

    @pytest.mark.parametrize(
        "connected, token, header, expected_connected, expected_token,"
        "expected_header",
        [
            # Case 1: Connection is active
            (
                True,
                "test_token",
                {"user-agent": "test-agent"},
                False,
                None,
                None,
            ),
            # Case 2: Connection is already inactive
            (
                False,
                "test_token",
                {"user-agent": "test-agent"},
                False,
                "test_token",
                {"user-agent": "test-agent"},
            ),
        ],
    )
    def test_reset_connection(
        self,
        connected: bool,
        token: Optional[str],
        header: Optional[dict[str, str]],
        expected_connected: bool,
        expected_token: Optional[str],
        expected_header: Optional[dict[str, str]],
    ) -> None:
        """Test the `reset_connection` method."""

        # Mock the `_new_session` method
        with patch.object(
            Connection, "_new_session", new_callable=Mock
        ) as mock_new_session:
            # Create a Connection
            connection = Connection(
                hostname="localhost",
                username="user",
                password="pass",
            )

            # Set the initial state
            connection._connected = connected
            connection._token = token
            connection._header = header

            # Call `reset_connection`
            connection.reset_connection()

            # Verify the state after calling `reset_connection`
            assert connection._connected == expected_connected
            assert connection._token == expected_token
            assert connection._header == expected_header

            # Verify `_new_session` was called during initialization
            mock_new_session.assert_called_once()

    @pytest.mark.parametrize(
        "initial_connected, expected_connected",
        [
            # Case 1: Connection is active
            (True, True),
            # Case 2: Connection is inactive
            (False, False),
        ],
    )
    def test_connected_property(
        self, initial_connected: bool, expected_connected: bool
    ) -> None:
        """Test the `connected` property."""

        # Mock the `_new_session` method
        with patch.object(
            Connection, "_new_session", new_callable=Mock
        ) as mock_new_session:
            # Create a Connection
            connection = Connection(
                hostname="localhost",
                username="user",
                password="pass",
            )

            # Set the initial state of `_connected`
            connection._connected = initial_connected

            # Verify the `connected` property
            assert connection.connected == expected_connected

            # Verify `_new_session` was called during initialization
            mock_new_session.assert_called_once()



"""Temperature endpoint module."""

from __future__ import annotations

import logging
import threading
from typing import Any, Optional

from asusrouter.config import ARConfig
from asusrouter.modules.data import AsusData
from asusrouter.modules.wlan import Wlan
from asusrouter.tools.cleaners import clean_content
from asusrouter.tools.converters import safe_float
from asusrouter.tools.readers import read_js_variables

_LOGGER = logging.getLogger(__name__)

EXPECTED_DECIMAL_PLACES = 2
EXPECTED_TEMPERATURE_MIN = 10.0
EXPECTED_TEMPERATURE_MAX = 150.0
MAX_SCALE_COUNT = 5

_temperature_warned: bool = False
_temperature_warned_lock = threading.Lock()


# The temperature data can be presented it the following JS variables:
# 1) curr_coreTmp_2_raw, curr_coreTmp_5_raw, curr_coreTmp_52_raw
# 2) curr_coreTmp_0_raw, curr_coreTmp_1_raw, curr_coreTmp_2_raw,
# curr_coreTmp_3_raw
# 3) curr_coreTmp_wl0_raw, curr_coreTmp_wl1_raw, curr_coreTmp_wl2_raw,
# curr_coreTmp_wl3_raw
# for 2ghz, 5ghz, 5ghz2, 6ghz respectively
# CPU temperature is set either in curr_cpuTemp or curr_coreTmp_cpu
def read(content: str) -> dict[str, Any]:
    """Read temperature data."""

    global _temperature_warned

    temperature: dict[str, Any] = {}

    # Prepare the content. This page is a set of JS variables
    variables = read_js_variables(clean_content(content))

    # Read WLAN temperatures
    # If there is curr_coreTmp_5_raw, we have type 1
    if "curr_coreTmp_5_raw" in variables:
        temperature[Wlan.FREQ_2G] = variables.get("curr_coreTmp_2_raw")
        temperature[Wlan.FREQ_5G] = variables.get("curr_coreTmp_5_raw")
        temperature[Wlan.FREQ_5G2] = variables.get("curr_coreTmp_52_raw")
    # If there is curr_coreTmp_0_raw, we have type 2
    elif "curr_coreTmp_0_raw" in variables:
        temperature[Wlan.FREQ_2G] = variables.get("curr_coreTmp_0_raw")
        temperature[Wlan.FREQ_5G] = variables.get("curr_coreTmp_1_raw")
        temperature[Wlan.FREQ_5G2] = variables.get("curr_coreTmp_2_raw")
        temperature[Wlan.FREQ_6G] = variables.get("curr_coreTmp_3_raw")
    # If there is curr_coreTmp_wl0_raw, we have type 3
    elif "curr_coreTmp_wl0_raw" in variables:
        temperature[Wlan.FREQ_2G] = variables.get("curr_coreTmp_wl0_raw")
        temperature[Wlan.FREQ_5G] = variables.get("curr_coreTmp_wl1_raw")
        temperature[Wlan.FREQ_5G2] = variables.get("curr_coreTmp_wl2_raw")
        temperature[Wlan.FREQ_6G] = variables.get("curr_coreTmp_wl3_raw")

    # Clean the temperature values from `&deg;C`
    for key, value in temperature.items():
        if value:
            temperature[key] = value.replace("&deg;C", "")

    # Read CPU temperature
    # If there is curr_coreTmp_cpu
    if "curr_coreTmp_cpu" in variables:
        temperature["cpu"] = variables.get("curr_coreTmp_cpu")
    # If there is curr_cpuTemp
    elif "curr_cpuTemp" in variables:
        temperature["cpu"] = variables.get("curr_cpuTemp")

    # Convert the temperature values to float or remove them
    # if they have "disabled"
    temperature = {
        key: safe_float(value)
        for key, value in temperature.items()
        if value and "disabled" not in value
    }

    # While this functional is performing a kind of post-processing,
    # it should stay a part of the read method to have access to the raw data
    if ARConfig.optimistic_temperature is True:
        temperature, scaled = _scale_temperature(temperature)
        with _temperature_warned_lock:
            if scaled and _temperature_warned is False:
                _LOGGER.warning(
                    "Temperature values were rescaled due to the issue with "
                    "the raw data. Please report this. The original data is: "
                    "`%s` and the expected range is between %s and %s.",
                    variables,
                    EXPECTED_TEMPERATURE_MIN,
                    EXPECTED_TEMPERATURE_MAX,
                )
                _temperature_warned = True

    return temperature


def process(data: dict[str, Any]) -> dict[AsusData, Any]:
    """Process temperature data."""

    temperature: dict[AsusData, Any] = {
        AsusData.TEMPERATURE: data,
    }

    return temperature


def _scale_temperature(
    temperature: dict[str, Optional[float]],
    range_min: float = EXPECTED_TEMPERATURE_MIN,
    range_max: float = EXPECTED_TEMPERATURE_MAX,
    max_scale_count: int = MAX_SCALE_COUNT,
) -> tuple[dict[str, float], bool]:
    """Scale temperature values to a range.

    This method is a temporary solution for those routers
    with orders of magnitude lower temperature values."""

    scaled_temperature = {}
    scaled = False

    for key, temp in temperature.items():
        if temp is None:
            continue

        original_temp = temp

        count = 0
        # Scale up if too small
        while temp < range_min and count < max_scale_count:
            temp *= 10
            count += 1

        # Scale down if too large
        while temp > range_max and count < max_scale_count:
            temp /= 10
            count += 1

        if temp != original_temp:
            scaled = True
        scaled_temperature[key] = round(temp, EXPECTED_DECIMAL_PLACES)

    return scaled_temperature, scaled



"""Update Clients endpoint module."""

from __future__ import annotations

import logging
import re
from json import JSONDecodeError
from typing import Any, Optional
from urllib import parse

from asusrouter.modules.data import AsusData
from asusrouter.tools.cleaners import clean_dict, clean_dict_key
from asusrouter.tools.readers import merge_dicts, read_json_content, readable_mac

_LOGGER = logging.getLogger(__name__)


LEGACY_WLAN: dict[str, int] = {
    "2g": 1,
    "5g": 2,
    "5g_2": 3,
}


def read(content: str) -> dict[str, Any]:
    """Read update clients data"""

    update_clients: dict[str, Any] = {}

    # Our JSON data is between `originData =` and `networkmap_fullscan = `
    regex = re.compile(r"originData = (.*)networkmap_fullscan = ")
    match = regex.search(content.replace("\n", ""))
    if match:
        # Old firmwares have a different format
        if "staticList" in match.group(1):
            _LOGGER.debug("Reading legacy update clients data")
            return read_legacy(match.group(1))
        # Modern firmware
        _LOGGER.debug("Reading modern update clients data")
        content = (
            match.group(1)
            .replace("fromNetworkmapd", '"fromNetworkmapd"')
            .replace("nmpClient :", '"nmpClient" :')
        )

        # Read the json content
        update_clients = read_json_content(content)

    return update_clients


def read_legacy(content: str) -> dict[str, Any]:
    """Read update clients data from legacy firmwares."""

    clients: dict[str, Any] = {}

    data = read_legacy_as_json(content)

    # Networkmap
    read_legacy_networmap(clients, data.get("fromNetworkmapd"))
    # Staticlist
    read_legacy_staticlist(clients, data.get("staticList"))
    # Wlan list
    read_legacy_wlan_list(clients, data)
    # Wlan info
    read_legacy_wlan_info(clients, data)
    # nmpClient
    read_legacy_nmpclient(clients, data.get("nmpClient"))

    # Just so that we don't need to process it anyhow special
    # This will be compatible with the new format
    return {"fromNetworkmapd": [clients]}


def read_legacy_as_json(content: str) -> dict[str, Any]:
    """Read update clients data from legacy firmwares as JSON."""

    # Step 0 - replace all `'` with `"`
    data = content.replace("'", '"')

    # Step 1 - remove all the replace and split statements
    to_replace = [
        '.replace(/&#62/g, ">")',
        '.replace(/&#60/g, "<")',
        '.split("<")',
        '.split(">")',
    ]
    for replace in to_replace:
        data = data.replace(replace, "")

    # Step 2 - remove the `decodeURIComponent` statements
    data = data.replace("decodeURIComponent(", "").replace(")", "")

    # Step 3 - Remove the `{}` in the beginning and end, also clean trailing and leading spaces
    data = data[1:-1].strip()

    # Step 4 - decode
    data = parse.unquote(data)

    # Step 5 - replace the keys by the regex
    # Find all `, [a-zA-Z0-9_]+:` where [a-zA-Z0-9_]+ is the key and add `"` around the key
    data = re.sub(r", ?([a-zA-Z0-9_]+):", r',"\1":', data)

    # Step 6 - replace `customList` with `"customList"`
    data = data.replace("customList", '"customList"')

    # Step 7 - find `"wlList_2g": \[[.*]\]` as well as `"wlList_5g": \[[.*]\]` and `"wlList_5g_2": \[[.*]\]`
    # and replace them with `"wlList_2g": {[.*]}` as well as `"wlList_5g": {[.*]}` and `"wlList_5g_2": {[.*]}`
    # Keep the content inside the brackets as is
    # Stop the regex search on the first occurence of the `]` so that we don't match the last `]`
    data = re.sub(r'"wlList_2g": \[([^\]]*)\]', r'"wlList_2g": {\1}', data, count=1)
    data = re.sub(r'"wlList_5g": \[([^\]]*)\]', r'"wlList_5g": {\1}', data, count=1)
    data = re.sub(r'"wlList_5g_2": \[([^\]]*)\]', r'"wlList_5g_2": {\1}', data, count=1)

    data = "{" + data + "}"
    try:
        return read_json_content(data)
    except JSONDecodeError:
        _LOGGER.debug("Failed to read legacy update clients data")
        return {}


def read_legacy_networmap(output: dict[str, Any], networkmap: Optional[str]) -> None:
    """Read legacy networkmap to the output dict."""

    if not networkmap:
        return

    client_list = networkmap.split("<")
    for client_line in client_list:
        if client_line == "":
            continue
        client = client_line.split(">")
        if len(client) < 4 or not readable_mac(client[3].upper()):
            continue
        mac = client[3].upper()
        output[mac] = {
            "mac": mac,
            "name": client[1] if len(client) > 1 else None,
            "ip": client[2] if len(client) > 2 else None,
            "ipMethod": "dhcp",
            "isWL": 0,
            "isOnline": 1,
        }


def read_legacy_nmpclient(output: dict[str, Any], nmp_client: Optional[str]) -> None:
    """Read legacy nmpclient to the output dict."""

    if not nmp_client:
        return

    client_list = nmp_client.split("<")
    for client_info in client_list:
        # Skip empty strings
        if client_info == "":
            continue
        # Split the string into a list by `>`
        client = client_info.split(">")
        # Check if client has enough elements
        if len(client) < 5:
            continue
        # 0 element is the MAC address written in lowercase and without `:`
        # We need to convert it to uppercase and add `:`
        mac = ":".join([client[0][i : i + 2].upper() for i in range(0, 12, 2)])
        client_type = client[4] if len(client) > 4 else None
        if mac in output:
            output[mac]["type"] = client_type
            continue
        output[mac] = {
            "mac": mac,
            "name": client[2] if len(client) > 2 else None,
            "type": client_type,
            "isOnline": 0,
        }


def read_legacy_staticlist(output: dict[str, Any], static_list: Optional[str]) -> None:
    """Read legacy staticlist to the output dict."""

    if not static_list:
        return

    client_list = static_list.split("<")
    for client_line in client_list:
        if client_line == "":
            continue
        client = client_line.split(">")
        mac = client[0].upper()
        if not readable_mac(mac):
            continue
        if mac in output:
            output[mac]["ipMethod"] = "static"
            continue
        output[mac] = {
            "mac": mac,
            "name": client[2] if len(client) > 2 else None,
            "ip": client[1] if len(client) > 1 else None,
            "ipMethod": "static",
            "isWL": 0,
            "isOnline": 0,
        }


def read_legacy_wlan_info(output: dict[str, Any], data: dict[str, Any]) -> None:
    """Read legacy wlan info to the output dict."""

    for wlan in LEGACY_WLAN:
        wlan_info = data.get(f"wlListInfo_{wlan}")
        if not wlan_info:
            continue

        for client in wlan_info:
            mac = client[0].upper()
            if not readable_mac(mac):
                continue
            curTx = client[1] if len(client) > 1 else None
            curRx = client[2] if len(client) > 2 else None
            wlConnectTime = client[3] if len(client) > 3 else None
            if mac in output:
                output[mac]["curTx"] = curTx
                output[mac]["curRx"] = curRx
                output[mac]["wlConnectTime"] = wlConnectTime
                continue
            output[mac] = {
                "mac": mac,
                "curTx": curTx,
                "curRx": curRx,
                "wlConnectTime": wlConnectTime,
            }


def read_legacy_wlan_list(output: dict[str, Any], data: dict[str, Any]) -> None:
    """Read legacy wlan list to the output dict."""

    for wlan, wid in LEGACY_WLAN.items():
        wlan_list = data.get(f"wlList_{wlan}")
        if not wlan_list:
            continue

        for mac, client in wlan_list.items():
            if not readable_mac(mac):
                continue
            rssi = client.get("rssi")
            if mac in output:
                output[mac]["isWL"] = wid
                output[mac]["rssi"] = rssi
                output[mac]["isOnline"] = 1
                continue
            output[mac] = {
                "mac": mac,
                "isWL": wid,
                "rssi": rssi,
                "isOnline": 1,
            }


def process(data: dict[str, Any]) -> dict[AsusData, Any]:
    """Process the update clients data."""

    state: dict[AsusData, Any] = {}

    # Clients from Networkmapd
    clients = {
        mac: description
        for mac, description in data.get("fromNetworkmapd", [{}])[0].items()
        if readable_mac(mac)
    }

    # Clients from nmpClient
    clients_historic = {
        mac: description
        for mac, description in data.get("nmpClient", [{}])[0].items()
        if readable_mac(mac)
    }

    # Merge states
    clients = merge_dicts(clients, clients_historic)
    # Clean the clients
    clean_dict(clients)
    # Clean the clients from the `from` field
    clients = clean_dict_key(clients, "from")

    state[AsusData.CLIENTS] = clients

    return state



"""Connection tools for AsusRouter."""

from __future__ import annotations

import aiohttp


def get_cookie_jar() -> aiohttp.CookieJar:
    """Get a cookie jar for the AsusRouter connection."""

    return aiohttp.CookieJar(unsafe=True, quote_cookie=False)



"""Onboarding endpoint module."""

from __future__ import annotations

from typing import Any

from asusrouter.modules.aimesh import AiMeshDevice
from asusrouter.modules.data import AsusData
from asusrouter.tools.cleaners import clean_content
from asusrouter.tools.converters import safe_bool, safe_int, safe_return
from asusrouter.tools.readers import read_json_content

CONNECTION_TYPE = {
    "2G": 1,
    "5G": 2,
    "5G1": 3,
    "5G2": 3,
    "6G": 4,
}


CONST_AP = {
    "2g": "2ghz",
    "5g": "5ghz",
    "5g1": "5ghz2",
    "6g": "6ghz",
    "dwb": "dwb",
}

CONST_FREQ = [
    "2g",
    "5g",
    "6g",
]


def read(content: str) -> dict[str, Any]:
    """Read onboarding data"""

    # Preprocess the content
    content = read_preprocess(content)

    # Read the json content
    onboarding: dict[str, Any] = read_json_content(content)

    return onboarding


def read_preprocess(content: str) -> str:
    """Preprocess the content before reading it."""

    # Prepare the content
    content = (
        clean_content(content)
        .replace(" = ", '":')
        .replace(";\n", ',"')
        .replace("[0]", "")
    )
    content = '{"' + content[:-3] + "}"

    # In case we have a trailing comma inside a dict
    content = content.replace(",}", "}")

    return content


def process(data: dict[str, Any]) -> dict[AsusData, Any]:
    """Process the onboarding data."""

    state: dict[AsusData, Any] = {}

    # AiMesh nodes state
    state[AsusData.AIMESH] = {
        node.mac: node
        for device in data.get("get_cfg_clientlist", [[]])[0]
        for node in [process_aimesh_node(device)]
    }

    # Client list
    clients = {}
    client_list = data.get("get_allclientlist", [{}])[0]
    for node in client_list:
        for connection in client_list[node]:
            convert = process_connection(connection)
            for mac in client_list[node][connection]:
                description = {
                    "connection_type": convert.get("connection_type"),
                    "guest": convert.get("guest"),
                    "ip": safe_return(
                        client_list[node][connection][mac].get("ip", None)
                    ),
                    "mac": mac,
                    "node": node,
                    "online": True,
                    "rssi": client_list[node][connection][mac].get(
                        "rssi", None
                    ),
                }
                clients[mac] = description

    state[AsusData.CLIENTS] = clients

    return state


def process_aimesh_node(data: dict[str, Any]) -> AiMeshDevice:
    """Process AiMesh node data."""

    # Get the list of WLAN APs
    ap = {}
    for el, value in CONST_AP.items():
        if f"ap{el}" in data and data[f"ap{el}"] is not str():
            ap[value] = data[f"ap{el}"]

    # Get the parent data
    parent = {}
    for el in CONST_FREQ:
        if f"pap{el}" in data and data[f"pap{el}"] is not str():
            parent["connection"] = CONST_AP[el]
            parent["mac"] = data[f"pap{el}"]
            parent["rssi"] = safe_return(data.get(f"rssi{el}"))
            parent["ssid"] = safe_return(data.get(f"pap{el}_ssid"))

            # Stop the loop since we found the parent
            break

    level = safe_int(data.get("level", "0"))
    node_type = "router" if level == 0 else "node"

    return AiMeshDevice(
        status=safe_bool(data.get("online", 0)) or False,
        alias=data.get("alias", None),
        model=data.get("ui_model_name", data.get("model_name", None)),
        product_id=data.get("product_id"),
        ip=data.get("ip"),
        fw=data.get("fwver", None),
        fw_new=safe_return(data.get("newfwver")),
        mac=data.get("mac", None),
        ap=ap,
        parent=parent,
        type=node_type,
        level=level,
        config=data.get("config"),
    )


def process_connection(data: str) -> dict[str, int]:
    """Process connection data."""

    # Check that the data is not empty
    if not isinstance(data, str) or data == "":
        return {}

    if data == "wired_mac":
        return {"connection_type": 0, "guest": 0}

    temp = data.split("_")
    return {
        "connection_type": CONNECTION_TYPE.get(temp[0]) or 0,
        "guest": safe_int(temp[1], 0) if len(temp) > 1 else 0,
    }



"""Aura module."""

from __future__ import annotations

import logging
import re
from enum import Enum, IntEnum
from typing import Any, Awaitable, Callable, Optional

from asusrouter.modules.color import (
    ColorRGB,
    ColorRGBB,
    average_color,
    parse_colors,
)
from asusrouter.modules.data import AsusData, AsusDataState
from asusrouter.modules.endpoint import EndpointTools
from asusrouter.modules.identity import AsusDevice
from asusrouter.tools.converters import (
    get_arguments,
    safe_bool,
    safe_enum,
    safe_int,
)

_LOGGER = logging.getLogger(__name__)


class AsusAura(IntEnum):
    """Asus Aura state."""

    UNKNOWN = -999
    ON = -1
    OFF = 0
    GRADIENT = 1
    STATIC = 2
    BREATHING = 3
    EVOLUTION = 4
    RAINBOW = 5
    WAVE = 6
    MARQUEE = 7


class AsusAuraColor(IntEnum):
    """Asus Aura with color support."""

    GRADIENT = 1
    STATIC = 2
    BREATHING = 3
    MARQUEE = 7


class DefaultAuraColor(Enum):
    """Default Aura colors."""

    COLOR1 = ColorRGBB((20, 0, 128))
    COLOR2 = ColorRGBB((110, 0, 100))
    COLOR3 = ColorRGBB((128, 0, 80))


DEFAULT_COLOR_PATTERN = (
    DefaultAuraColor.COLOR1,
    DefaultAuraColor.COLOR2,
    DefaultAuraColor.COLOR3,
    DefaultAuraColor.COLOR2,
)

DEFAULT_AURA_SCHEME = AsusAura.STATIC


def get_default_aura_color(zones: int) -> tuple[ColorRGBB, ...]:
    """Get the default Aura color for the zones."""

    _pattern_length = len(DEFAULT_COLOR_PATTERN)
    # Repeat the pattern enough times to cover the required length
    repeated_pattern = DEFAULT_COLOR_PATTERN * (
        (zones + _pattern_length - 1) // _pattern_length
    )

    # Return the required length
    return tuple(color.value for color in repeated_pattern[:zones])


def get_scheme_from_state(aura_state: dict) -> AsusAura:
    """Get the scheme from the state."""

    def is_valid_scheme(scheme: Any) -> bool:
        """Check if the scheme is valid."""

        # Keep compatible with Python 3.11,
        valid_schemes = [
            e.value
            for e in AsusAura
            if e not in (AsusAura.UNKNOWN, AsusAura.ON, AsusAura.OFF)
        ]
        return scheme in valid_schemes

    current_scheme = safe_int(aura_state.get("scheme"))
    if current_scheme is not None and is_valid_scheme(current_scheme):
        return AsusAura(current_scheme)

    prev_scheme = safe_int(aura_state.get("scheme_prev"))
    if prev_scheme is not None and is_valid_scheme(prev_scheme):
        return AsusAura(prev_scheme)

    return DEFAULT_AURA_SCHEME


def set_color(
    colors: list[ColorRGBB],
    color_to_set: Optional[ColorRGB | list[ColorRGB]],
    zone: Optional[int] = None,
    zones: Optional[int] = None,
) -> None:
    """Set new color(s) to the existing color list."""

    # Fallback in case zones have wrong data written
    zones = zones or len(colors)

    # Make sure, we have enough colors in the list (fix wrong data)
    if len(colors) < zones:
        colors.extend(colors[: zones - len(colors)])

    # No new color defined
    if not color_to_set:
        _LOGGER.debug("No new color defined. Skipping")
        return

    if isinstance(color_to_set, ColorRGB):
        _LOGGER.debug("Single color defined: %s", color_to_set)
        # Single zone defined
        if zone is not None and zone < zones:
            _LOGGER.debug("Applying the color to the zone %s", zone)
            colors[zone].from_rgb(color_to_set)
        # No zone defined - set the color to all zones
        else:
            _LOGGER.debug("Applying the color to all zones")
            for i in range(zones):
                colors[i].from_rgb(color_to_set)
        return

    if isinstance(color_to_set, list):
        # We have a list of colors to set
        _LOGGER.debug("List of colors defined")
        for i in range(zones):
            color = color_to_set[i % len(color_to_set)]
            if isinstance(color, ColorRGB):
                colors[i].from_rgb(color)

    return


def set_brightness(
    colors: list[ColorRGBB],
    brightness: Optional[int],
    zone: Optional[int] = None,
) -> None:
    """Set the brightness for the zones."""

    # Prepare the brightness
    brightness = safe_int(brightness)

    # No brightness defined
    if brightness is None:
        return

    # Change the brightness for the selected zone
    if isinstance(zone, int) and 0 <= zone < len(colors):
        colors[zone].set_brightness(brightness)
        return

    # Change the brightness for all zones
    for color in colors:
        color.set_brightness(brightness)

    return


async def set_state(
    callback: Callable[..., Awaitable[bool]],
    state: AsusAura,
    **kwargs: Any,
) -> bool:
    """Set the Aura state."""

    # Get the identity
    identity: AsusDevice = kwargs.get("identity", AsusDevice())
    # Get the number of zones
    zones = identity.aura_zone
    if zones < 1:
        _LOGGER.debug("No Aura zones found. Skipping the Aura service.")
        return False

    # Get the current router state
    aura_state = (
        kwargs.get("router_state", {}).get(AsusData.AURA, AsusDataState).data
    )

    # Get the arguments
    color_to_set, brightness, zone = get_arguments(
        ("color", "brightness", "zone"), **kwargs
    )

    if state == AsusAura.ON:
        _LOGGER.debug("Aura effect not specified. Checking the previous state")
        # Get the previous state
        state = get_scheme_from_state(aura_state)

    # Check if the state has color support
    if state.name not in AsusAuraColor.__members__:
        # No color selection for the state
        _LOGGER.debug(
            "Setting the Aura state to `%s`. No color / brightness support for this state",
            state.name,
        )
        return await callback(
            endpoint=EndpointTools.AURA,
            commands={"ledg_scheme": state.value},
        )

    # Get the state number
    state_number = AsusAuraColor[state.name].value

    # Get the previous colors / fallback to the default color
    colors = aura_state.get("effect", {}).get(
        state_number, get_default_aura_color(zones)
    )
    # Convert to ColorRGBB if needed
    colors = [
        ColorRGBB(color) if isinstance(color, ColorRGB) else color
        for color in colors
    ]

    # Set new color(s) and brightness
    set_color(colors, color_to_set, zone, zones)
    set_brightness(colors, brightness, zone)

    # Convert color_set to the string
    color_to_use = ",".join(
        [color_zone.to_rgb().__str__() for color_zone in colors]
    )
    _LOGGER.debug("Setting the Aura color to `%s`", color_to_use)

    # Prepare the arguments
    arguments = {
        "ledg_scheme": state.value,
        "ledg_rgb": color_to_use,
    }

    # Run the service
    return await callback(
        endpoint=EndpointTools.AURA,
        commands=arguments,
    )


def process_aura(data: dict[str, Any]) -> dict[str, Any]:
    """Process Aura data."""

    def get_scheme(key: str) -> AsusAura:
        """Get the scheme from the data."""

        scheme_value = data.get(key)
        _scheme_value = safe_int(scheme_value)
        _scheme_enum = safe_enum(AsusAura, _scheme_value, default_value=-999)
        scheme: AsusAura = (
            _scheme_enum if _scheme_enum is not None else AsusAura.UNKNOWN
        )

        if scheme == AsusAura.UNKNOWN and _scheme_value is not None:
            _LOGGER.warning("Unknown Aura scheme: `%s`", scheme_value)
        return scheme

    # Get the effects data
    rgb_key_pattern = re.compile(r"ledg_rgb(\d+)")
    effect = {
        safe_int(match.group(1)): parse_colors(data[key])
        for key in data
        if (match := rgb_key_pattern.match(key))
    }

    aura = {
        "state": safe_bool(data.get("AllLED")),
        "night_mode": safe_bool(data.get("ledg_night_mode")),
        "scheme": get_scheme("ledg_scheme"),
        "scheme_prev": get_scheme("ledg_scheme_old"),
        "effect": effect,
        "active": {},
    }

    if aura["scheme"] == AsusAura.OFF:
        aura["state"] = False

    # Active effect
    active_effect = aura["effect"].get(aura["scheme"].value)
    if active_effect:
        _active_brightness = 0
        for i, color in enumerate(active_effect):
            _brightness = color.brightness
            aura["active"][i] = {"color": color, "brightness": _brightness}
            _active_brightness = max(_active_brightness, _brightness)

        # Active color and brightness
        active_color = ColorRGBB(average_color(active_effect))
        active_color.set_brightness(_active_brightness)
        aura["active"]["color"] = active_color
        aura["active"]["brightness"] = _active_brightness

    print(aura["effect"])

    # Get number of zones from the Static effect length
    _effect_static = aura["effect"].get(AsusAuraColor.STATIC)
    aura["zones"] = len(_effect_static) if _effect_static else 0

    return aura



"""OpenVPN module for AsusRouter."""

from __future__ import annotations

import logging
from enum import IntEnum
from typing import Any, Awaitable, Callable

from asusrouter.modules.firmware import Firmware
from asusrouter.tools.converters import get_arguments

_LOGGER = logging.getLogger(__name__)

REQUIRE_IDENTITY = True


class AsusOVPNClient(IntEnum):
    """Asus OpenVPN client state."""

    UNKNOWN = -999
    ERROR = -1
    DISCONNECTED = 0
    CONNECTING = 1
    CONNECTED = 2

    OFF = DISCONNECTED
    ON = CONNECTING


class AsusOVPNServer(IntEnum):
    """Asus OpenVPN server state."""

    UNKNOWN = -999
    ERROR = -1
    DISCONNECTED = 0
    CONNECTING = 1
    CONNECTED = 2

    OFF = DISCONNECTED
    ON = CONNECTING


async def set_state(
    callback: Callable[..., Awaitable[bool]],
    state: AsusOVPNClient | AsusOVPNServer,
    **kwargs: Any,
) -> bool:
    """Set the OpenVPN state."""

    # Check if state is available
    if not isinstance(state, (AsusOVPNClient, AsusOVPNServer)) or not state.value in (
        0,
        1,
    ):
        _LOGGER.debug("No state found in arguments")
        return False

    # Get the arguments
    vpn_id, identity = get_arguments(("id", "identity"), **kwargs)

    if not vpn_id:
        _LOGGER.debug("No VPN id found in arguments")
        return False

    service_arguments = {"id": vpn_id}

    service_map: dict[Any, str]

    # Get the correct service call
    # This will be firmware dependent
    if (
        not identity
        or identity.merlin
        or identity.firmware < Firmware(major="3.0.0.4", minor=388, build=0)
    ):
        service_map = {
            (AsusOVPNClient, AsusOVPNClient.ON): f"start_vpnclient{vpn_id}",
            (AsusOVPNClient, AsusOVPNClient.OFF): f"stop_vpnclient{vpn_id}",
            (AsusOVPNServer, AsusOVPNServer.ON): f"start_vpnserver{vpn_id}",
            (AsusOVPNServer, AsusOVPNServer.OFF): f"stop_vpnserver{vpn_id}",
        }
        service = service_map.get((type(state), state))
    else:
        service_map = {
            AsusOVPNServer.ON: (
                "restart_openvpnd;restart_chpass;restart_samba;restart_dnsmasq;"
            ),
            AsusOVPNServer.OFF: "stop_openvpnd;restart_samba;restart_dnsmasq;",
        }
        service = service_map.get(state) if isinstance(state, AsusOVPNServer) else None
        service_arguments["VPNServer_enable"] = (
            "1" if state == AsusOVPNServer.ON else "0"
        )

    if not service:
        _LOGGER.debug("Unknown state %s", state)
        return False

    _LOGGER.debug(
        "Triggering state set with parameters: service=%s, arguments=%s",
        service,
        service_arguments,
    )

    # Run the service
    return await callback(
        service=service,
        arguments=service_arguments,
        apply=True,
        expect_modify=kwargs.get("expect_modify", False),
    )



"""Access Point module."""

from __future__ import annotations

from dataclasses import dataclass


@dataclass
class AccessPoint:
    """Access point class."""

    # Basic information
    mac: str | None = None
    ssid: str | None = None

    # Fronthaul information
    hidden: bool | None = None
    mac_fh: str | None = None
    ssid_fh: str | None = None



"""VPNC module. This module is compatible with VPN Fusion in AsusWRT 388+."""

from __future__ import annotations

import logging
from enum import Enum, IntEnum
from typing import Any, Awaitable, Callable, Optional

from asusrouter.modules.data import AsusData, AsusDataState
from asusrouter.modules.openvpn import AsusOVPNClient
from asusrouter.modules.wireguard import AsusWireGuardClient
from asusrouter.tools.converters import get_arguments

_LOGGER = logging.getLogger(__name__)


REQUIRE_STATE = True


class AsusVPNC(IntEnum):
    """Asus VPN Fusion state."""

    UNKNOWN = -999
    CONNECTING = 1
    CONNECTED = 2
    ERROR = 4
    DISCONNECTED = 5

    OFF = DISCONNECTED
    ON = CONNECTING


class AsusVPNType(str, Enum):
    """Asus VPN Fusion type."""

    UNKNOWN = "Unknown"
    L2TP = "L2TP"
    OPENVPN = "OpenVPN"
    PPTP = "PPTP"
    SURFSHARK = "Surfshark"
    WIREGUARD = "WireGuard"


async def set_state(
    callback: Callable[..., Awaitable[bool]],
    state: AsusVPNC | AsusOVPNClient | AsusWireGuardClient,
    **kwargs: Any,
) -> bool:
    """Set the VPN state."""

    # Match the state to call the correct function
    match state:
        case a if isinstance(a, AsusVPNC):
            return await set_state_vpnc(callback, a, **kwargs)
        case a if isinstance(a, (AsusOVPNClient, AsusWireGuardClient)):
            return await set_state_other(callback, a, **kwargs)
        case _:
            try:
                _LOGGER.debug(
                    "Unknown state %s. Cannot find proper handler", state.name
                )
            except AttributeError:
                _LOGGER.debug("Unknown state %s. Cannot find proper handler", state)
            return False


VPNC_STATE_MAPPING = {
    AsusVPNC.ON: ("restart_vpnc", 1),
    AsusVPNC.OFF: ("stop_vpnc", 0),
}


async def set_state_vpnc(
    callback: Callable[..., Awaitable[bool]],
    state: Optional[AsusVPNC],
    **kwargs: Any,
) -> bool:
    """Set the VPN Fusion state."""

    # Check if state is available
    if not isinstance(state, AsusVPNC):
        _LOGGER.debug("No state found in arguments")
        return False

    # Get the arguments
    vpnc_unit = get_arguments("vpnc_unit", **kwargs)

    if not isinstance(vpnc_unit, int):
        _LOGGER.debug("No VPN Fusion unit found in arguments")
        return False

    # Service arguments
    service_arguments: dict[str, Any] = {"vpnc_unit": vpnc_unit}

    # Get the raw state from router
    router_state = kwargs.get("router_state", {})

    # Clientlist is needed to update the state (all clients are in the list).
    # If not available, we cannot update the state.
    vpnc_clientlist = router_state.get(AsusData.VPNC_CLIENTLIST, AsusDataState()).data
    if not vpnc_clientlist or vpnc_clientlist == "":
        _LOGGER.debug("No VPN Fusion client list found in router state")
        return False

    # Get service and binary state
    service, binary_state = VPNC_STATE_MAPPING.get(state, (None, None))
    if service is None:
        _LOGGER.debug("Unknown state %s", state)
        return False

    # Update the clientlist
    vpnc_clientlist = _get_argument_clientlist(vpnc_clientlist, vpnc_unit, binary_state)
    if not vpnc_clientlist:
        _LOGGER.debug("Something went wrong with creating a new clientlist")
        return False

    service_arguments["vpnc_clientlist"] = vpnc_clientlist

    _LOGGER.debug(
        "Triggering state set with parameters: service=%s, arguments=%s",
        service,
        service_arguments,
    )

    # Call the service
    return await callback(
        service=service,
        arguments=service_arguments,
        apply=True,
        expect_modify=kwargs.get("expect_modify", False),
    )


OTHER_STATE_MAPPING = {
    AsusOVPNClient.ON: AsusVPNC.ON,
    AsusOVPNClient.OFF: AsusVPNC.OFF,
    AsusWireGuardClient.ON: AsusVPNC.ON,
    AsusWireGuardClient.OFF: AsusVPNC.OFF,
}


async def set_state_other(
    callback: Callable[..., Awaitable[bool]],
    state: AsusOVPNClient | AsusWireGuardClient,
    **kwargs: Any,
) -> bool:
    """Set the Open VPN state."""

    # Check if state is available
    if not isinstance(
        state, (AsusOVPNClient, AsusWireGuardClient)
    ) or not state.value in (0, 1):
        _LOGGER.debug("No state found in arguments")
        return False

    # Get the arguments
    vpn_id = get_arguments("id", **kwargs)

    if not isinstance(vpn_id, int):
        _LOGGER.debug("No VPN id found in arguments")
        return False

    # Get the raw state from router
    router_state = kwargs.get("router_state", {})

    # Clientlist is needed to update the state (all clients are in the list).
    # If not available, we cannot update the state.
    vpnc_data = router_state.get(AsusData.VPNC, AsusDataState()).data

    vpnc_unit = _find_vpnc_unit(
        vpnc_data=vpnc_data, client_type=state, client_id=vpn_id
    )

    # Convert state
    vpnc_state = OTHER_STATE_MAPPING.get(state)

    return await set_state_vpnc(
        callback=callback,
        state=vpnc_state,
        vpnc_unit=vpnc_unit,
        expect_modify=kwargs.get("expect_modify", False),
        router_state=router_state,
    )


def _get_argument_clientlist(
    clientlist: Optional[str], vpnc_unit: Optional[int], state: Optional[int]
) -> Optional[str]:
    """Generate the clientlist argument."""

    if not clientlist or clientlist == "" or vpnc_unit is None or state is None:
        return None

    # Split clientlist properly
    clients: list[str] = clientlist.split("<")
    if len(clients) <= vpnc_unit:
        return None
    client = clients[vpnc_unit]

    # In the client data get the 6th parameter which is the state
    client_param = client.split(">")
    client_param[5] = str(state)

    # Assemble client
    client = ">".join(client_param)
    # Assemble clientlist
    clients[vpnc_unit] = client
    clientlist = "<".join(clients)

    return clientlist


def _find_vpnc_unit(
    vpnc_data: Optional[dict[AsusVPNType, dict[int, dict[str, Any]]]],
    client_type: AsusOVPNClient | AsusWireGuardClient,
    client_id: int,
) -> Optional[int]:
    """Find vpnc unit by VPN client_id."""

    if vpnc_data is None:
        return None

    match client_type:
        case a if isinstance(a, AsusOVPNClient):
            search_type = AsusVPNType.OPENVPN
        case a if isinstance(a, AsusWireGuardClient):
            search_type = AsusVPNType.WIREGUARD
        case _:
            _LOGGER.debug("Unknown client type %s", client_type)
            return None

    if search_type not in vpnc_data:
        return None

    if client_id not in vpnc_data[search_type]:
        return None

    return vpnc_data[search_type][client_id].get("vpnc_unit")



"""Configuration module for AsusRouter."""

from __future__ import annotations

import threading
from typing import Any, Callable, Dict, Optional

from asusrouter.tools.converters import safe_bool

CONFIG_DEFAULT_BOOL: bool = False


def safe_bool_config(value: Any) -> bool:
    """Convert a value to a boolean, defaulting to CONFIG_DEFAULT_BOOL."""

    config_value: Optional[bool] = safe_bool(value)

    if config_value is None:
        return CONFIG_DEFAULT_BOOL

    return config_value


class Config:
    """Configuration class for AsusRouter."""

    def __init__(self) -> None:
        """Initialize the configuration."""

        self._lock = threading.Lock()

        self._options: Dict[str, Any] = {
            "optimistic_data": CONFIG_DEFAULT_BOOL,
            "optimistic_temperature": CONFIG_DEFAULT_BOOL,
        }
        self._types: Dict[str, Callable[[Any], Any]] = {}

    def set(self, key: str, value: Any) -> None:
        """Set the configuration option."""

        with self._lock:
            if key in self._options:
                converter: Callable[[Any], Any] = self._types.get(
                    key, safe_bool_config
                )
                self._options[key] = converter(value)
            else:
                raise KeyError(f"Unknown configuration option: {key}")

    def get(self, key: str) -> Any:
        """Get the configuration option."""

        with self._lock:
            if key in self._options:
                return self._options[key]
            else:
                raise KeyError(f"Unknown configuration option: {key}")

    def keys(self) -> list[str]:
        """Get the list of configuration keys."""

        with self._lock:
            return list(self._options.keys())

    @property
    def optimistic_data(self) -> bool:
        """Get the optimistic data flag."""

        return bool(self.get("optimistic_data"))

    @property
    def optimistic_temperature(self) -> bool:
        """Get the optimistic temperature flag."""

        return bool(self.get("optimistic_temperature"))


ARConfig: Config = Config()



"""Writers module.

This module contains the writers for AsusRouter, which prepare data to be sent to the router."""


from __future__ import annotations

from asusrouter.tools.converters import clean_input


@clean_input
def nvram(content: str | list[str] | None = None) -> str | None:
    """NVRAM writer.

    This function converts a list of strings (or a single string)
    into a string request to the NVRAM read endpoint."""

    if isinstance(content, str):
        return f"nvram_get({content});"

    if isinstance(content, list):
        return "".join([f"nvram_get({item});" for item in content])

    return None



"""Connection module.

This module handles the connection between the library and the router
as well as all the data transfer."""

from __future__ import annotations

import asyncio
import base64
import json
import logging
from typing import Any, Awaitable, Callable, Optional, Tuple
from urllib.parse import quote

import aiohttp

from asusrouter.const import DEFAULT_TIMEOUT, USER_AGENT, RequestType
from asusrouter.error import (
    AsusRouter404Error,
    AsusRouterAccessError,
    AsusRouterConnectionError,
    AsusRouterError,
    AsusRouterLogoutError,
    AsusRouterTimeoutError,
)
from asusrouter.modules.endpoint import EndpointService, EndpointType
from asusrouter.modules.endpoint.error import handle_access_error
from asusrouter.tools.connection import get_cookie_jar

_LOGGER = logging.getLogger(__name__)


def generate_credentials(
    username: str, password: str
) -> Tuple[str, dict[str, str]]:
    """Generate credentials for connection"""

    auth = f"{username}:{password}".encode("ascii")
    logintoken = base64.b64encode(auth).decode("ascii")
    payload = f"login_authorization={logintoken}"
    headers = {"user-agent": USER_AGENT}

    return payload, headers


class Connection:  # pylint: disable=too-many-instance-attributes
    """A connection between the library and the device."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        hostname: str,
        username: str,
        password: str,
        port: Optional[int] = None,
        use_ssl: bool = False,
        session: Optional[aiohttp.ClientSession] = None,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
        dumpback: Optional[Callable[..., Awaitable[None]]] = None,
    ):
        """Initialize connection."""

        _LOGGER.debug("Initializing a new connection to `%s`", hostname)

        # Initialize parameters for connection
        self._token: Optional[str] = None
        self._header: Optional[dict[str, str]] = None
        self._connected: bool = False
        self._connection_lock: asyncio.Lock = asyncio.Lock()
        self._timeout: int = timeout or DEFAULT_TIMEOUT

        # Hostname and credentials
        self._hostname = hostname
        self._username = username
        self._password = password

        # Set the port and protocol based on the input
        self._http = "https" if use_ssl else "http"
        self._port = port or (8443 if use_ssl else 80)
        self._ssl = use_ssl
        self._verify_ssl = False
        _LOGGER.debug(
            "Using `%s` and port `%s` with ssl flag `%s`",
            self._http,
            self._port,
            self._ssl,
        )

        # Callback for dumping data
        self._dumpback = dumpback

        # Client session
        self._manage_session: bool = False
        if session is not None:
            _LOGGER.debug("Using provided session")
            self._session = session
        else:
            _LOGGER.debug("No session provided. Will create a new one")
            self._session = self._new_session()

    async def __aenter__(self) -> Connection:
        """Enter the connection."""

        await self.async_connect()
        return self

    async def __aexit__(
        self,
        exc_type: Any,
        exc_value: Any,
        traceback: Any,
    ) -> None:
        """Exit the connection."""

        await self.async_close()

    @classmethod
    async def create(
        cls,
        hostname: str,
        username: str,
        password: str,
        port: Optional[int] = None,
        use_ssl: bool = False,
        session: Optional[aiohttp.ClientSession] = None,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
        dumpback: Optional[Callable[..., Awaitable[None]]] = None,
    ) -> Connection:
        """Factory method to create and initialize a connection."""

        connection = cls(
            hostname=hostname,
            username=username,
            password=password,
            port=port,
            use_ssl=use_ssl,
            session=session,
            timeout=timeout or DEFAULT_TIMEOUT,
            dumpback=dumpback,
        )
        await connection.async_connect()
        return connection

    def _new_session(self) -> aiohttp.ClientSession:
        """Create a new session."""

        # If we create a new session, we will manage it
        self._manage_session = True

        # Timeout for the session
        timeout = aiohttp.ClientTimeout(total=self._timeout)

        # Create the session
        return aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(ssl=self._verify_ssl),
            cookie_jar=get_cookie_jar(),
            timeout=timeout,
        )

    async def async_close(self) -> None:
        """Close the session."""

        if self._manage_session and self._session:
            if not self._session.closed:
                _LOGGER.debug("Closing the session")
                await self._session.close()
            else:
                _LOGGER.debug("Session already closed")
        else:
            _LOGGER.debug("No session to close or not managing the session")

    async def async_connect(self, lock: Optional[asyncio.Lock] = None) -> bool:
        """Connect to the device and get a new auth token."""

        try:
            await asyncio.wait_for(
                self._async_connect_with_lock(lock), timeout=self._timeout
            )
            return True
        except asyncio.TimeoutError:
            _LOGGER.error("Connection to %s timed out", self._hostname)
            return False

    async def _async_connect_with_lock(
        self,
        lock: Optional[asyncio.Lock] = None,
    ) -> bool:
        """Connect to the device and get a new auth token."""

        # Check that we are connected
        # so that we don't try to go through lock again
        if self._connected:
            _LOGGER.debug("Already connected to %s", self._hostname)
            return True

        _lock = lock or self._connection_lock

        async with _lock:
            _LOGGER.debug("Initializing connection to %s", self._hostname)
            self._connected = False

            # Generate payload and header for login request
            payload, headers = generate_credentials(
                self._username, self._password
            )

            # Request authotization
            _LOGGER.debug("Requesting authorization")
            try:
                _, _, resp_content = await self._send_request(
                    EndpointService.LOGIN, payload, headers
                )
                _LOGGER.debug("Received authorization response")
            except AsusRouterAccessError as ex:
                raise AsusRouterAccessError(
                    f"Cannot access {EndpointService.LOGIN}. "
                    "Failed in `async_connect`"
                ) from ex
            except AsusRouterError as ex:
                _LOGGER.debug("Connection failed with error: %s", ex)
                raise
            except Exception as ex:  # pylint: disable=broad-except
                _LOGGER.debug(
                    "Unexpected error while connecting to %s: %s",
                    self._hostname,
                    ex,
                )
                raise ex

            # Convert the response to JSON
            content = json.loads(resp_content)
            # Get the auth_token value from the headers
            self._token = content.get("asus_token")
            if not self._token:
                _LOGGER.error("No token received")
                return False

            # Set the header
            self._header = {
                "user-agent": USER_AGENT,
                "cookie": f"asus_token={self._token}",
            }

            # Mark as connected
            self._connected = True
            _LOGGER.debug("Connected to %s", self._hostname)

            return True

    async def async_disconnect(self) -> bool:
        """Disconnect from the device."""

        _LOGGER.debug("Initializing disconnection from %s", self._hostname)

        # Check that we are connected
        if not self._connected:
            _LOGGER.debug("Not connected to %s", self._hostname)
            return True

        # Request logout
        try:
            await self._send_request(EndpointService.LOGOUT)
        except AsusRouterLogoutError:
            # Loged out successfully
            self.reset_connection()
            _LOGGER.debug("Disconnected from %s", self._hostname)
            return True
        except AsusRouterError as ex:
            _LOGGER.debug(
                "Error while disconnecting from %s: %s", self._hostname, ex
            )
            return False

        # Anything else would mean error when disconnecting
        return False

    async def _send_request(
        self,
        endpoint: EndpointType,
        payload: Optional[str] = None,
        headers: Optional[dict[str, str]] = None,
        request_type: RequestType = RequestType.POST,
    ) -> Tuple[int, dict[str, str], str]:
        """Send a request to the device."""

        # Send request
        try:
            resp_status, resp_headers, resp_content = await self._make_request(
                endpoint,
                payload,
                headers,
                request_type,
            )

            # Raise exception on 404
            if resp_status == 404:
                raise AsusRouter404Error(f"Endpoint {endpoint} not found")

            # Raise exception on non-200 status
            if resp_status != 200:
                raise AsusRouterAccessError(
                    f"Cannot access {endpoint}, status {resp_status}"
                )

            # Check for access errors
            if "error_status" in resp_content:
                handle_access_error(
                    endpoint, resp_status, resp_headers, resp_content
                )

            # Return the response
            return (resp_status, resp_headers, resp_content)
        except aiohttp.ClientConnectorError as ex:
            self.reset_connection()
            raise AsusRouterConnectionError(
                f"Cannot connect to `{self._hostname}` on port "
                f"`{self._port}`. Failed in `_send_request` with error: `{ex}`"
            ) from ex
        except (aiohttp.ClientConnectionError, aiohttp.ClientOSError) as ex:
            raise AsusRouterConnectionError(
                f"Cannot connect to `{self._hostname}` on port "
                f"`{self._port}`. Failed in `_send_request` with error: `{ex}`"
            ) from ex
        except (asyncio.TimeoutError, asyncio.CancelledError) as ex:
            raise AsusRouterTimeoutError(
                f"Data cannot be retrieved due to an asyncio error. "
                f"Connection failed: {ex}"
            ) from ex
        except Exception as ex:  # pylint: disable=broad-except
            _LOGGER.debug(
                "Unexpected error sending request to %s: %s",
                endpoint,
                ex,
            )
            raise ex

    async def async_query(
        self,
        endpoint: EndpointType,
        payload: Optional[str] = None,
        headers: Optional[dict[str, str]] = None,
        request_type: RequestType = RequestType.POST,
    ) -> Tuple[int, dict[str, str], str]:
        """Send a request to the device."""

        # If not connected, try to connect
        if not self._connected:
            _LOGGER.debug("Not connected to %s. Connecting...", self._hostname)
            await self.async_connect()

        # If still not connected, raise an error
        if not self._connected:
            raise AsusRouterTimeoutError(
                "Data cannot be retrieved. Connection failed"
            )

        # Send the request
        _LOGGER.debug(
            "Sending `%s` request to `%s`", request_type, self._hostname
        )
        return await self._send_request(
            endpoint, payload, headers, request_type
        )

    async def _make_request(
        self,
        endpoint: EndpointType,
        payload: Optional[str] = None,
        headers: Optional[dict[str, str]] = None,
        request_type: RequestType = RequestType.POST,
    ) -> Any:
        """Make a post request to the device."""

        # Check if a session is available
        if self._session is None or self._session.closed:
            # If no session is available, we cannot be connected to the device
            _LOGGER.debug("No session available. Creating a new one")
            # We will create a new session and retry the request
            self.reset_connection()
            self._session = self._new_session()
            # Reconnect
            await self.async_connect()
            # Retry the request
            return await self._make_request(
                endpoint, payload, headers, request_type
            )

        # Check headers
        if not headers:
            headers = self._header

        # Generate the url
        url = f"{self._http}://{self._hostname}:{self._port}/{endpoint.value}"

        # Add get parameters if needed
        if request_type == RequestType.GET and payload:
            payload.replace(";", "&")
            url = f"{url}?{payload}"

        # Process the payload to be sent
        payload_to_send = quote(payload) if payload else None

        # Send the request
        async with self._session.request(
            request_type.value,
            url,
            data=payload_to_send if request_type == RequestType.POST else None,
            headers=headers,
        ) as response:
            # Read the status code
            resp_status = response.status

            # Read the response headers
            resp_headers = response.headers

            # Read the response
            try:
                resp_content = await response.text()
            except UnicodeDecodeError:
                _LOGGER.debug("Cannot decode response. Will ignore errors")
                resp_content = await response.text(errors="ignore")

            # Call the dumpback if available
            if self._dumpback is not None:
                await self._dumpback(
                    endpoint, payload, resp_status, resp_headers, resp_content
                )

            # Return the response
            return (resp_status, resp_headers, resp_content)

    def reset_connection(self) -> None:
        """Reset connection variables."""

        if not self._connected:
            return

        _LOGGER.debug("Resetting connection to %s", self._hostname)

        self._connected = False
        self._token = None
        self._header = None

    @property
    def connected(self) -> bool:
        """Return connection status."""

        return self._connected



"""Data finder module."""

from __future__ import annotations

import logging
from enum import Enum
from typing import Callable, Optional

from asusrouter.modules.attributes import AsusRouterAttribute
from asusrouter.modules.data import AsusData
from asusrouter.modules.endpoint import Endpoint, EndpointTools, EndpointType
from asusrouter.modules.endpoint.hook_const import (
    MAP_OVPN_SERVER_388,
    MAP_VPNC_WIREGUARD,
    MAP_WAN,
    MAP_WAN_ITEM,
    MAP_WAN_ITEM_X,
    MAP_WIREGUARD_CLIENT,
    MAP_WIREGUARD_SERVER,
)
from asusrouter.modules.parental_control import HOOK_PC
from asusrouter.modules.wlan import gwlan_nvram_request, wlan_nvram_request
from asusrouter.tools import converters

_LOGGER = logging.getLogger(__name__)


class AsusDataMerge(str, Enum):
    """AsusRouter data merge class."""

    ALL = "all"
    ANY = "any"


class AsusDataFinder:
    """AsusRouter data finder class."""

    def __init__(
        self,
        endpoint: list[EndpointType] | EndpointType,
        merge: AsusDataMerge = AsusDataMerge.ANY,
        request: Optional[list[tuple[str, ...]]] = None,
        nvram: Optional[list[str] | str] = None,
        method: Optional[Callable] = None,
        arguments: Optional[AsusRouterAttribute] = None,
    ) -> None:
        """Initialize the data finder."""

        # Set the endpoint as list even if it's a single endpoint
        if not isinstance(endpoint, list):
            endpoint = [endpoint]
        self.endpoint = endpoint

        # Set the merge
        self.merge = merge

        # Set the request and append nvram hooks to the request
        self.request = request or []
        if nvram:
            nvram_request = converters.nvram_get(nvram)
            if nvram_request:
                self.request.extend(nvram_request)

        # Set the method and arguments
        self.method = method
        self.arguments = arguments


# A constant list of requests for fetching data
ASUSDATA_REQUEST = {
    "devices": [
        ("get_clientlist", ""),
    ],
    "main": [
        ("cpu_usage", "appobj"),
        ("memory_usage", "appobj"),
        ("netdev", "appobj"),
    ],
    "speedtest": [
        ("ookla_speedtest_get_result", ""),
    ],
    # "speedtest_history": [
    #     ("ookla_speedtest_get_history", ""),
    # ],
    # "speedtest_servers": [
    #     ("ookla_speedtest_get_servers", ""),
    # ],
    "vpnc": [
        ("get_vpnc_status", ""),
    ],
    "wan": [
        ("get_wan_unit", ""),
    ],
    "wireguard_server": [
        ("get_wgsc_status", ""),
    ],
}

ASUSDATA_NVRAM = {
    "aura": [
        "AllLED",
        "ledg_night_mode",
        "ledg_scheme",
        "ledg_scheme_old",
    ],
    "light": ["led_val"],
    "openvpn_server_388": [
        key
        for element in MAP_OVPN_SERVER_388
        for key, _, _ in [converters.safe_unpack_keys(element)]
    ],
    "parental_control": HOOK_PC,
    "port_forwarding": [
        "vts_rulelist",
        "vts_enable_x",
    ],
    "speedtest": ["ookla_state"],
    "vpnc": [
        "vpnc_clientlist",
    ],
    "wan": [
        key
        for element in MAP_WAN
        for key, _, _ in [converters.safe_unpack_keys(element)]
        if key != "get_wan_unit"
    ],
    "wireguard_server": [
        key
        for element in MAP_WIREGUARD_SERVER
        for key, _, _ in [converters.safe_unpack_keys(element)]
        if key != "get_wgsc_status"
    ],
    "dsl": [
        "dsllog_dataratedown",
        "dsllog_datarateup",
    ]
}
ASUSDATA_NVRAM["aura"].extend([f"ledg_rgb{num}" for num in range(0, 8)])
ASUSDATA_NVRAM["vpnc"].extend(
    [
        f"wgc{num}_{key}"
        for num in range(1, 6)
        for element in MAP_VPNC_WIREGUARD
        for key, _, _ in [converters.safe_unpack_keys(element)]
    ]
)
ASUSDATA_NVRAM["wan"].extend(
    [
        f"wan{num}_{key}"
        for num in (0, 1)
        for element in MAP_WAN_ITEM
        for key, _, _ in [converters.safe_unpack_keys(element)]
    ]
)
ASUSDATA_NVRAM["wan"].extend(
    [
        f"wan{num}_{extra}{key}"
        for num in (0, 1)
        for extra in ("", "x")
        for element in MAP_WAN_ITEM_X
        for key, _, _ in [converters.safe_unpack_keys(element)]
    ]
)
ASUSDATA_NVRAM["wireguard_server"].extend(
    [
        f"wgs1_c{num}_{key}"
        for num in range(1, 11)
        for element in MAP_WIREGUARD_CLIENT
        for key, _, _ in [converters.safe_unpack_keys(element)]
    ]
)

ASUSDATA_ENDPOINT_APPEND = {
    Endpoint.PORT_STATUS: {
        # Request status of the ports for the whole AiMesh network
        # This will save time and requests, since we then cache the data
        # in most cases
        "node_mac": "all",
    }
}


# A map of endptoins to get data from
ASUSDATA_MAP: dict[AsusData, AsusData | AsusDataFinder] = {
    AsusData.AIMESH: AsusDataFinder(Endpoint.ONBOARDING),
    AsusData.AURA: AsusDataFinder(
        Endpoint.HOOK,
        nvram=ASUSDATA_NVRAM["aura"],
    ),
    AsusData.BOOTTIME: AsusData.DEVICEMAP,
    AsusData.CLIENTS: AsusDataFinder(
        [Endpoint.ONBOARDING, Endpoint.UPDATE_CLIENTS], AsusDataMerge.ALL
    ),
    AsusData.CPU: AsusDataFinder(
        Endpoint.HOOK, request=ASUSDATA_REQUEST["main"]
    ),
    AsusData.DEVICEMAP: AsusDataFinder(Endpoint.DEVICEMAP),
    AsusData.FIRMWARE: AsusDataFinder(Endpoint.FIRMWARE),
    AsusData.FIRMWARE_NOTE: AsusDataFinder(
        [Endpoint.FIRMWARE_NOTE, Endpoint.FIRMWARE_NOTE_AIMESH]
    ),
    AsusData.GWLAN: AsusDataFinder(
        Endpoint.HOOK,
        method=gwlan_nvram_request,
        arguments=AsusRouterAttribute.WLAN_LIST,
    ),
    AsusData.LED: AsusDataFinder(Endpoint.HOOK, nvram=ASUSDATA_NVRAM["light"]),
    AsusData.NETWORK: AsusData.CPU,
    AsusData.NODE_INFO: AsusDataFinder(Endpoint.PORT_STATUS),
    AsusData.OPENVPN: AsusDataFinder(
        [Endpoint.VPN, Endpoint.DEVICEMAP], AsusDataMerge.ANY
    ),
    AsusData.OPENVPN_CLIENT: AsusData.OPENVPN,
    AsusData.OPENVPN_SERVER: AsusData.OPENVPN,
    AsusData.PARENTAL_CONTROL: AsusDataFinder(
        Endpoint.HOOK, nvram=ASUSDATA_NVRAM["parental_control"]
    ),
    AsusData.PING: AsusDataFinder(EndpointTools.NETWORK),
    AsusData.PORT_FORWARDING: AsusDataFinder(
        Endpoint.HOOK, nvram=ASUSDATA_NVRAM["port_forwarding"]
    ),
    AsusData.PORTS: AsusDataFinder(
        [Endpoint.PORT_STATUS, Endpoint.ETHERNET_PORTS]
    ),
    AsusData.RAM: AsusData.CPU,
    AsusData.SPEEDTEST: AsusDataFinder(
        Endpoint.HOOK,
        nvram=ASUSDATA_NVRAM["speedtest"],
        request=ASUSDATA_REQUEST["speedtest"],
    ),
    # AsusData.SPEEDTEST_HISTORY: AsusDataFinder(
    #     Endpoint.HOOK, request=ASUSDATA_REQUEST["speedtest_history"]
    # ),
    AsusData.SPEEDTEST_RESULT: AsusData.SPEEDTEST,
    # AsusData.SPEEDTEST_SERVERS: AsusDataFinder(
    #     Endpoint.HOOK, request=ASUSDATA_REQUEST["speedtest_servers"]
    # ),
    AsusData.SYSINFO: AsusDataFinder(Endpoint.SYSINFO),
    AsusData.TEMPERATURE: AsusDataFinder(Endpoint.TEMPERATURE),
    AsusData.VPNC: AsusDataFinder(
        Endpoint.HOOK,
        nvram=ASUSDATA_NVRAM["vpnc"],
        request=ASUSDATA_REQUEST["vpnc"],
    ),
    AsusData.VPNC_CLIENTLIST: AsusData.VPNC,
    AsusData.WAN: AsusDataFinder(
        Endpoint.HOOK,
        nvram=ASUSDATA_NVRAM["wan"],
        request=ASUSDATA_REQUEST["wan"],
    ),
    AsusData.WIREGUARD: AsusData.WIREGUARD_SERVER,
    AsusData.WIREGUARD_CLIENT: AsusData.VPNC,
    AsusData.WIREGUARD_SERVER: AsusDataFinder(
        Endpoint.HOOK,
        nvram=ASUSDATA_NVRAM["wireguard_server"],
        request=ASUSDATA_REQUEST["wireguard_server"],
    ),
    AsusData.WLAN: AsusDataFinder(
        Endpoint.HOOK,
        method=wlan_nvram_request,
        arguments=AsusRouterAttribute.WLAN_LIST,
    ),
    AsusData.DSL: AsusDataFinder(
        Endpoint.HOOK,
        nvram=ASUSDATA_NVRAM["dsl"],
    ),
}


def add_conditional_data_rule(data: AsusData, rule: AsusDataFinder) -> None:
    """A callback to add / change rule for ASUSDATA_MAP."""

    ASUSDATA_MAP[data] = rule
    _LOGGER.debug("Added conditional data rule: %s -> %s", data, rule)


def add_conditional_data_alias(data: AsusData, origin: AsusData) -> None:
    """A callback to add / change rule for ASUSDATA_MAP."""

    ASUSDATA_MAP[data] = origin
    _LOGGER.debug("Added data alias: %s -> %s", origin, data)


def remove_data_rule(data: AsusData) -> None:
    """A callback to remove rule for ASUSDATA_MAP."""

    ASUSDATA_MAP.pop(data, None)
    _LOGGER.debug("Removed data rule: %s", data)



"""System module for AsusRouter."""

from __future__ import annotations

import logging
from enum import Enum
from typing import Any, Awaitable, Callable

_LOGGER = logging.getLogger(__name__)

ACTION_MODE = "action_mode"
APPLY = "apply"
ARGUMENTS = "arguments"
EXPECT_MODIFY = "expect_modify"
SERVICE = "service"


class AsusSystem(str, Enum):
    """Asus system enum.

    This enum contains all known system services. The availability of these
    services depends on the router model and firmware version.

    Some services might not be tested and might not work as expected. Use with
    caution and at your own risk."""

    # ---------------------
    # DISCLAIMER: Migration to the new format
    #
    # AsusSystem enum is being migrated to a new format. Instead of
    # `{ACTION}_{FEATURE}`, the format should be `{FEATURE}_{ACTION}`. This
    # will allow better grouping of related actions and easier access.
    #
    # E.g. `REBUILD_AIMESH` -> `AIMESH_REBUILD`
    #
    # The old format will be deprecated and removed in a future version.
    # Please, only add new actions in the new format.
    #
    # All the old actions will be migrated to the new format with time
    # (and updates in the corresponding modules). Any deprecated actions
    # with a new format should be added to the `AsusSystemDeprecated` dict.
    #
    # Check the AiMesh for an example of the new format.
    # ---------------------

    # ---------------------
    # AiMesh
    AIMESH_ADD_NODE = "ob_selection"  # Add a new node to AiMesh
    AIMESH_ONBOARDING = (
        "onboarding"  # Start AiMesh onboarding (search for nodes)
    )
    AIMESH_REBOOT = "device_reboot"  # Restart router + all nodes
    AIMESH_REBUILD = "re_reconnect"  # Rebuild AiMesh
    NODE_CONFIG_CHANGE = "config_changed"  # Change node configuration
    NODE_REBOOT = "node_reboot"  # Reboot a node by MAC (when specified)
    REBUILD_AIMESH = "_depr_re_reconnect"  # Rebuild AiMesh / legacy name
    # ---------------------
    # Aura
    AURA_RESTART = "restart_ledg"  # Restart Aura RGB
    # ---------------------
    # Firmware
    FIRMWARE_CHECK = "firmware_check"  # Check for firmware update
    # Firmware upgrade will upgrade the firmware to the latest version
    # if available. The firmware file is downloaded from the Asus server.
    # After the download, the router will reboot and install the firmware.
    FIRMWARE_UPGRADE = "firmware_upgrade"  # Firmware upgrade
    # ---------------------
    # Upgrades and updates
    UPGRADE_RESTART = "restart_upgrade"  # Upgrade firmware and restart device
    UPGRADE_START = "start_upgrade"  # Start firmware upgrade / installation
    UPGRADE_STOP = "stop_upgrade"  # Stop firmware upgrade / installation
    RESTART_UPGRADE = "_depr_restart_upgrade"
    START_UPGRADE = "_depr_start_upgrade"
    STOP_UPGRADE = "_depr_stop_upgrade"
    # ---------------------
    PREPARE_CERT = "prepare_cert"  # Prepare certificate
    REBOOT = "reboot"  # Reboot the router
    RESTART_CHPASS = "restart_chpass"
    RESTART_CLOUDSYNC = "restart_cloudsync"  # AiCloud 2.0 sync
    RESTART_CP = "restart_CP"  # Captive Portal
    RESTART_DDNS_LE = "restart_ddns_le"
    RESTART_DEFAULT_WAN = "restart_default_wan"  # Default WAN
    RESTART_DISKMON = "restart_diskmon"  # Disk monitor
    RESTART_DNSFILTER = "restart_dnsfilter"  # DNS filter
    RESTART_DNSMASQ = "restart_dnsmasq"  # DNS
    RESTART_DNSQD = "restart_dnsqd"  #
    RESTART_FIREWALL = "restart_firewall"  # Firewall
    RESTART_FTPD = "restart_ftpd"  # FTP server
    RESTART_FTPSAMBA = "restart_ftpsamba"
    RESTART_HTTPD = "restart_httpd"  # Web server
    RESTART_KEY_GUARD = "restart_key_guard"
    RESTART_LEDS = "restart_leds"
    RESTART_LPD = "restart_lpd"  # Line Printer Daemon (LPD)
    RESTART_NASAPPS = "restart_nasapps"
    RESTART_NET = "restart_net"  # Restart all wired interfaces
    RESTART_NET_AND_PHY = (
        "restart_net_and_phy"  # Restart all network interfaces
    )
    RESTART_NFCM = "restart_nfcm"
    RESTART_OAM = "restart_oam"
    RESTART_OPENVPND = "restart_openvpnd"  # OpenVPN daemon
    RESTART_QOS = "restart_qos"  # Quality of Service (QoS)
    RESTART_ROUTERBOOST = "restart_routerboost"
    RESTART_SAMBA = "restart_samba"  # Samba file sharing
    RESTART_SETTINGS_WEBDAV = "restart_settings_webdav"  # WebDAV
    RESTART_SNMPD = (
        "restart_snmpd"  # Simple Network Management Protocol (SNMP)
    )
    RESTART_SUBNET = "restart_subnet"
    RESTART_TIME = "restart_time"
    RESTART_TIMEMACHINE = "restart_timemachine"
    RESTART_TOR = "restart_tor"  # The Onion Router (TOR)
    RESTART_TR = "restart_tr"
    RESTART_U2EC = "restart_u2ec"  # USB to Ethernet Connector
    RESTART_UPNP = "restart_upnp"  # Universal Plug and Play (UPnP)
    RESTART_USB_IDLE = "restart_usb_idle"
    RESTART_VPNC = "restart_vpnc"
    RESTART_VPND = "restart_vpnd"  # VPN daemon for legacy VPNs
    RESTART_WAN = (
        "restart_wan"  # Restart WAN connection (optional if number, e.g. 0)
    )
    RESTART_WAN_DNS = "restart_wan_dns"  # Restart WAN DNS (optional if number)
    RESTART_WAN_IF = (
        "restart_wan_if"  # Restart WAN interfaces (optional if number)
    )
    RESTART_WEBDAV = "restart_webdav"  # WebDAV
    RESTART_WIRELESS = "restart_wireless"  # Restart all wireless interfaces
    RESTART_WLCSCAN = "restart_wlcscan"
    RESTART_WGS = "restart_wgs"  # Restart all wireguard server interfaces
    RESTART_WPSIE = "restart_wpsie"
    RESTART_WRS = "restart_wrs"
    RESTART_WTFAST_RULE = "restart_wtfast_rule"
    RESET_LED = "reset_led"
    START_AURARGB = "start_aurargb"  # Aura RGB
    START_DISKFORMAT = "start_diskformat"  # Format disk
    START_DISKSCAN = "start_diskscan"  # Scan disk
    START_WEBS_UPGRADE = (
        "start_webs_upgrade"  # Start firmware upgrade from the web
    )
    START_WEBS_UPDATE = "start_webs_update"  # Check for firmware update
    START_WRS = "start_wrs"
    STOP_AURARGB = "stop_aurargb"  # Aura RGB
    STOP_LOGGER = "stop_logger"
    STOP_OPENVPND = "stop_openvpnd"  # Stop OpenVPN server
    STOP_VPNC = "stop_vpnc"
    STOP_VPND = "stop_vpnd"  # Stop VPN server for legacy VPNs
    STOP_WRS_FORCE = "stop_wrs_force"
    UPDATE_CLIENTS = "update_clients"


# Deprecated states for AsusSystem
# Format: {DeprecatedState: (NewState, Version)}
# Version is the version when the deprecated state will be removed (optional)
AsusSystemDeprecated = {
    # AiMesh
    AsusSystem.REBUILD_AIMESH: (AsusSystem.AIMESH_REBUILD, None),
    # Upgrades and updates
    AsusSystem.RESTART_UPGRADE: (AsusSystem.UPGRADE_RESTART, None),
    AsusSystem.START_UPGRADE: (AsusSystem.UPGRADE_START, None),
    AsusSystem.STOP_UPGRADE: (AsusSystem.UPGRADE_STOP, None),
}


# Map AsusSystem special cases to service calls
STATE_MAP: dict[AsusSystem, dict[str, Any]] = {
    AsusSystem.AIMESH_REBOOT: {
        SERVICE: None,
        ARGUMENTS: {ACTION_MODE: AsusSystem.AIMESH_REBOOT.value},
        APPLY: False,
        EXPECT_MODIFY: False,
    },
    AsusSystem.NODE_CONFIG_CHANGE: {
        SERVICE: None,
        ARGUMENTS: {ACTION_MODE: AsusSystem.NODE_CONFIG_CHANGE.value},
        APPLY: False,
        EXPECT_MODIFY: False,
    },
    AsusSystem.NODE_REBOOT: {
        SERVICE: None,
        # NODE_REBOOT is an alias for AIMESH_REBOOT, but it requires
        # passing a node MAC address. For the safety, this AsusSystem
        # state is separated to avoid accidental reboot of the whole
        # AiMesh network.
        ARGUMENTS: {ACTION_MODE: AsusSystem.AIMESH_REBOOT.value},
        APPLY: False,
        EXPECT_MODIFY: False,
    },
    AsusSystem.FIRMWARE_CHECK: {
        SERVICE: None,
        ARGUMENTS: {ACTION_MODE: AsusSystem.FIRMWARE_CHECK.value},
        APPLY: False,
        EXPECT_MODIFY: False,
    },
    AsusSystem.FIRMWARE_UPGRADE: {
        SERVICE: None,
        ARGUMENTS: {ACTION_MODE: AsusSystem.FIRMWARE_UPGRADE.value},
        APPLY: False,
        EXPECT_MODIFY: False,
    },
    AsusSystem.UPDATE_CLIENTS: {
        SERVICE: None,
        ARGUMENTS: {ACTION_MODE: AsusSystem.UPDATE_CLIENTS.value},
        APPLY: False,
        EXPECT_MODIFY: False,
    },
}

ARGUMENTS_APPEND: dict[AsusSystem, list[str]] = {
    AsusSystem.NODE_CONFIG_CHANGE: ["re_mac", "config"],
    AsusSystem.NODE_REBOOT: ["device_list"],
}


async def set_state(
    callback: Callable[..., Awaitable[bool]],
    state: AsusSystem,
    **kwargs: Any,
) -> bool:
    """Set the system state."""

    # Check and notify if the state is deprecated
    if state in AsusSystemDeprecated:
        repl_state, repl_ver = AsusSystemDeprecated[state]
        message = f"Deprecated state `{state.name}` from `AsusSystem` \
enum used. Use `{repl_state.name}` instead"
        if repl_ver:
            message += f". This state will be removed in version {repl_ver}"
        _LOGGER.warning(
            message,
        )
        state = repl_state

    arguments: dict[str, Any] = kwargs.get(ARGUMENTS, {})

    # Get the arguments for the callback function based on the state
    callback_args: dict[str, Any] = STATE_MAP.get(
        state,
        {
            SERVICE: state.value,
            ARGUMENTS: arguments,
            APPLY: True,
            EXPECT_MODIFY: kwargs.get(EXPECT_MODIFY, False),
        },
    )

    # Append additional arguments if the state is in the append list
    if state in ARGUMENTS_APPEND:
        search_args = ARGUMENTS_APPEND[state]
        if not all(arg in kwargs for arg in search_args):
            _LOGGER.error(
                "Setting state `%s` requires ALL of the following "
                "arguments: %s",
                state,
                search_args,
            )
            return False

        # Append the arguments to the callback arguments
        callback_args[ARGUMENTS].update(
            {arg: kwargs[arg] for arg in search_args}
        )

    # Run the service
    return await callback(**callback_args)



"""Test AsusRouter with real devices data."""

import importlib
import logging
import os
import re
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any, Dict, Optional

import pytest

from asusrouter import AsusData
from asusrouter.modules.endpoint import (
    Endpoint,
    EndpointControl,
    EndpointService,
    EndpointTools,
    EndpointType,
    process,
    read,
)

# Create a logger
_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.INFO)


class FileExtensions(Enum):
    CONTENT = ".content"
    PYTHON = ".py"


@dataclass
class DataItem:
    """A class used to represent a test item."""

    content: str
    result: Dict[AsusData, Any]
    endpoint: EndpointType
    label: str

    def __repr__(self):
        return self.label


@contextmanager
def log_test_data_loading():
    """Log the start and end of loading test data."""

    _LOGGER.info("Starting to load test data")
    yield
    _LOGGER.info("Finished loading test data")


def load_content_data(device_path: Path, module_name: str) -> str:
    """Load content data from a file."""

    content_file = device_path / f"{module_name}{FileExtensions.CONTENT.value}"
    with content_file.open("r", encoding="utf-8") as f:
        return f.read()


def load_expected_result(device_path: Path, module_name: str) -> Any:
    """Load expected result from a module."""

    result_module = importlib.import_module(
        f".test_data.{'.'.join([device_path.name, module_name])}",
        package="tests",
    )
    return result_module.expected_result


def load_test_item(device_path: Path, module_name: str) -> Optional[DataItem]:
    """Load a single test item."""

    try:
        endpoint_name = re.match(r"(.*)_\d+", module_name).group(1)

        endpoint = None
        for EndpointEnum in [Endpoint, EndpointControl, EndpointService, EndpointTools]:
            try:
                endpoint = EndpointEnum[endpoint_name.upper()]
                break
            except KeyError:
                continue

        if endpoint is None:
            raise ValueError("Failed to load test item ", module_name)

        content_data = load_content_data(device_path, module_name)
        expected_result = load_expected_result(device_path, module_name)

        return DataItem(
            content=content_data,
            result=expected_result,
            endpoint=endpoint,
            label=f"{device_path.name}_{module_name}",
        )
    except Exception as ex:  # pylint: disable=broad-except
        _LOGGER.error("Failed to load test item %s: %s", module_name, ex)
        return None


def load_test_data() -> list[DataItem]:
    """Load the test data."""

    with log_test_data_loading():
        test_data_path = Path(__file__).resolve().parent / "test_data"
        data = []

        for device_path in test_data_path.iterdir():
            if device_path.is_dir():
                device_test_count = 0
                for content_file in device_path.glob("*.content"):
                    module_name = content_file.stem

                    # Check if both .content and .py files exist
                    if (
                        not (device_path / f"{module_name}.content").exists()
                        or not (device_path / f"{module_name}.py").exists()
                    ):
                        continue

                    item = load_test_item(device_path, module_name)
                    data.append(item)
                    device_test_count += 1

                _LOGGER.info(
                    "Found %s test items for device: %s",
                    device_test_count,
                    device_path.name,
                )

        _LOGGER.info("Total test items found: %s", len(data))
        return data


# Load the test data only once
test_data = load_test_data()

# Create a list of ids for the test data
test_ids = [item.label for item in test_data]


@pytest.fixture(params=test_data, ids=test_ids)
def test_item(request):
    """Yield each item in the test data."""

    return request.param


def test_asusrouter(test_item: DataItem):  # pylint: disable=redefined-outer-name
    """
    Test the asusrouter module with the given test item.

    Args:
        test_item (DataItem): The test item to use for the test.

    Raises:
        AssertionError: If the actual processed data does not match the expected result.
    """

    actual_read = read(test_item.endpoint, test_item.content)
    actual_processed = process(test_item.endpoint, actual_read)

    assert actual_processed == test_item.result, print(actual_processed)



"""AsusRouter module.

This module contains the main class for interacting with an Asus device."""

from __future__ import annotations

import asyncio
import json
import logging
from datetime import datetime, timedelta, timezone
from typing import Any, Awaitable, Callable, Optional

import aiohttp

from asusrouter.connection import Connection
from asusrouter.const import (
    DEFAULT_CACHE_TIME,
    DEFAULT_RESULT_SUCCESS,
    DEFAULT_TIMEOUT,
    RequestType,
)
from asusrouter.error import (
    AsusRouter404Error,
    AsusRouterAccessError,
    AsusRouterConnectionError,
    AsusRouterDataError,
)
from asusrouter.modules.attributes import AsusRouterAttribute
from asusrouter.modules.data import AsusData, AsusDataState
from asusrouter.modules.data_finder import (
    ASUSDATA_ENDPOINT_APPEND,
    ASUSDATA_MAP,
    ASUSDATA_NVRAM,
    AsusDataFinder,
    AsusDataMerge,
    add_conditional_data_alias,
    add_conditional_data_rule,
    remove_data_rule,
)
from asusrouter.modules.data_transform import (
    transform_clients,
    transform_cpu,
    transform_ethernet_ports,
    transform_network,
    transform_wan,
)
from asusrouter.modules.endpoint import (
    ENDPOINT_FORCE_REQUEST,
    Endpoint,
    EndpointControl,
    EndpointType,
    process,
    read,
)
from asusrouter.modules.endpoint.error import AccessError
from asusrouter.modules.firmware import Firmware
from asusrouter.modules.flags import Flag
from asusrouter.modules.identity import AsusDevice, collect_identity
from asusrouter.modules.port_forwarding import PortForwardingRule
from asusrouter.modules.service import async_call_service
from asusrouter.modules.state import (
    AsusState,
    add_conditional_state,
    get_datatype,
    keep_state,
    save_state,
    set_state,
)
from asusrouter.tools import legacy
from asusrouter.tools.converters import get_enum_key_by_value, safe_list
from asusrouter.tools.readers import merge_dicts

_LOGGER = logging.getLogger(__name__)


class AsusRouter:
    """The interface class."""

    def __init__(
        self,
        hostname: str,
        username: str,
        password: str,
        port: Optional[int] = None,
        use_ssl: bool = False,
        cache_time: Optional[float] = None,
        session: Optional[aiohttp.ClientSession] = None,
        dumpback: Optional[Callable[..., Awaitable[None]]] = None,
    ):  # pylint: disable=too-many-arguments
        """Initialize the interface."""

        _LOGGER.debug("Initializing a new interface to `%s`", hostname)

        # Set the cache time
        self._cache_time = cache_time or DEFAULT_CACHE_TIME

        # Set the host
        self._hostname: str = hostname

        # Set the device identity
        self._identity: Optional[AsusDevice] = None
        self._state: dict[AsusData, AsusDataState] = {}

        # Set the flags
        self._flags: Flag = Flag()
        # Time for change to take effect before available to fetch
        self._needed_time: Optional[int] = None
        # ID from the last called service
        self._last_id: Optional[int] = None

        # Create an empty connection and save the credentials
        self._connection: Optional[Connection] = None
        self._username = username
        self._password = password
        self._port = port
        self._use_ssl = use_ssl
        self._session = session
        self._dumpback = dumpback

    # ---------------------------
    # Connection-related methods -->
    # ---------------------------

    async def async_init_connection(self) -> None:
        """Initialize the connection."""

        _LOGGER.debug("Triggered method async_init_connection")

        self._connection = await Connection.create(
            hostname=self._hostname,
            username=self._username,
            password=self._password,
            port=self._port,
            use_ssl=self._use_ssl,
            session=self._session,
            timeout=DEFAULT_TIMEOUT,
            dumpback=self._dumpback,
        )

    async def async_del_connection(self) -> None:
        """Delete the connection."""

        _LOGGER.debug("Triggered method async_del_connection")

        # Disconnect from the device if connected
        await self.async_disconnect()

        # Close the connection
        if self._connection:
            await self._connection.async_close()
            self._connection = None

    async def async_connect(self) -> bool:
        """Connect to the device and get its identity."""

        _LOGGER.debug("Triggered method async_connect")

        # Connect to the device
        try:
            # Make sure the connection is initialized
            if self._connection is None:
                await self.async_init_connection()

            # Connect to the device
            result = (
                await self._connection.async_connect()
                if self._connection
                else False
            )
            if result is False:
                return False
        except Exception as ex:  # pylint: disable=broad-except
            raise ex

        # Get the device identity
        return await self.async_get_identity() is not None

    async def async_disconnect(self) -> bool:
        """Disconnect from the device."""

        _LOGGER.debug("Triggered method async_disconnect")

        # Disconnect from the device
        try:
            if self._connection:
                await self._connection.async_disconnect()
        except Exception as ex:  # pylint: disable=broad-except
            await self._async_handle_exception(ex)

        return True

    async def _async_drop_connection(self) -> None:
        """Drop the connection in case we know it cannot reply
        due to our last actions."""

        _LOGGER.debug("Triggered method _async_drop_connection")

        if self._connection:
            self._connection.reset_connection()

    async def _async_handle_exception(self, ex: Exception) -> None:
        """Handle exceptions."""

        _LOGGER.debug("Triggered method _async_handle_exception")

        raise ex

    async def _async_handle_reboot(self) -> None:
        """Handle reboot."""

        _LOGGER.debug("Triggered method _async_handle_reboot")

        led_state = self._state.get(AsusData.LED)
        if led_state and led_state.data:
            _LOGGER.debug("Restoring LED state")
            await keep_state(
                callback=self.async_run_service,
                states=led_state.data["state"],
                identity=self._identity,
            )

        # Reset the reboot flag
        self._reset_flag("reboot")

    def _reset_flag(self, flag: str) -> None:
        """Reset a flag."""

        _LOGGER.debug("Triggered method _reset_flag")

        # Check that AsusData.FLAGS is available
        if AsusData.FLAGS not in self._state:
            return

        # Get the data from the state
        data = self._state[AsusData.FLAGS].data

        # Check that data is a dict
        if not isinstance(data, dict):
            return

        # Reset the flag
        data.pop(flag, None)

        _LOGGER.debug("Flag `%s` reset", flag)

    # ---------------------------
    # <-- Connection-related methods
    # ---------------------------

    # ---------------------------
    # Identity-related methods -->
    # ---------------------------

    async def async_get_identity(self, force: bool = False) -> AsusDevice:
        """Get the device identity."""

        _LOGGER.debug("Triggered method async_get_identity")

        # Check whether we already have the identity and not forcing a refresh
        if self._identity and not force:
            return self._identity

        # Collect the identity
        self._identity = await collect_identity(
            api_hook=self.async_api_hook,
            api_query=self.async_api_query,
        )

        # Add conditional data rules
        if self._identity:
            firmware = self._identity.firmware
            merlin = self._identity.merlin
            fw_388 = Firmware(major="3.0.0.4", minor=388, build=0)
            # Stock
            if not merlin:
                _LOGGER.debug("Adding conditional rules for stock firmware")
                if fw_388 < firmware:
                    add_conditional_state(
                        AsusState.OPENVPN_CLIENT, AsusData.VPNC
                    )
                    add_conditional_state(
                        AsusState.WIREGUARD_CLIENT, AsusData.VPNC
                    )
                    add_conditional_data_alias(
                        AsusData.OPENVPN_CLIENT, AsusData.VPNC
                    )
                    add_conditional_data_alias(
                        AsusData.WIREGUARD_CLIENT, AsusData.VPNC
                    )
                    add_conditional_data_rule(
                        AsusData.OPENVPN_SERVER,
                        AsusDataFinder(
                            Endpoint.HOOK,
                            nvram=ASUSDATA_NVRAM["openvpn_server_388"],
                        ),
                    )
            # Merlin
            else:
                _LOGGER.debug("Adding conditional rules for Merlin firmware")
                if fw_388 < firmware:
                    add_conditional_data_rule(
                        AsusData.VPNC,
                        AsusDataFinder(
                            Endpoint.HOOK,
                            nvram=ASUSDATA_NVRAM["vpnc"],
                        ),
                    )
            # Before 388
            if firmware < fw_388:
                # Remove VPNC rules
                remove_data_rule(AsusData.VPNC)
                remove_data_rule(AsusData.VPNC_CLIENTLIST)
                # Remove WireGuard rules
                remove_data_rule(AsusData.WIREGUARD)
                remove_data_rule(AsusData.WIREGUARD_CLIENT)
                remove_data_rule(AsusData.WIREGUARD_SERVER)

            # DSL connection
            if self._identity.dsl is False:
                remove_data_rule(AsusData.DSL)

            # Ookla Speedtest
            if self._identity.ookla is False:
                remove_data_rule(AsusData.SPEEDTEST)
                # remove_data_rule(AsusData.SPEEDTEST_HISTORY)
                remove_data_rule(AsusData.SPEEDTEST_RESULT)
                # remove_data_rule(AsusData.SPEEDTEST_SERVERS)

        # Return new identity
        return self._identity

    # ---------------------------
    # <-- Identity-related methods
    # ---------------------------

    # ---------------------------
    # Request-related methods -->
    # ---------------------------

    def _get_attribute(
        self, attribute: Optional[AsusRouterAttribute]
    ) -> Optional[Any]:
        """Get an attribute value."""

        if not attribute:
            return None

        match attribute:
            case AsusRouterAttribute.MAC:
                if self._identity:
                    return self._identity.mac
            case AsusRouterAttribute.WLAN_LIST:
                if self._identity:
                    return self._identity.wlan

        return None

    async def _check_flags(self) -> None:
        """Check flags."""

        _LOGGER.debug("Triggered method _check_flags")

        flags: dict[str, bool] = {}

        state_flags = self._state.get(AsusData.FLAGS)
        if isinstance(state_flags, AsusDataState) and isinstance(
            state_flags.data, dict
        ):
            flags = state_flags.data

        if flags.get("reboot", False) is True:
            _LOGGER.debug("Reboot flag is set")
            await self._async_handle_reboot()

    async def async_api_query(
        self, endpoint: EndpointType, payload: Optional[str] = None
    ):
        """A wrapper for the connection query method."""

        if endpoint in ASUSDATA_ENDPOINT_APPEND:
            payload = payload or ""
            for key, attribute in ASUSDATA_ENDPOINT_APPEND[endpoint].items():
                if isinstance(attribute, AsusRouterAttribute):
                    value = self._get_attribute(attribute)
                else:
                    value = attribute
                if value:
                    payload += f"{key}={value};"
            # Remove trailing semicolon
            payload = payload[:-1]

        _LOGGER.debug(
            "Triggered method async_api_query: %s | %s", endpoint, payload
        )

        request_type = ENDPOINT_FORCE_REQUEST.get(endpoint, RequestType.POST)

        return await self._connection.async_query(
            endpoint, payload, request_type=request_type
        )

    async def async_api_load(
        self,
        endpoint: EndpointType,
        request: str = "",
        retry: int = 0,
    ) -> dict[str, Any]:
        """Load API endpoint with optional request."""

        _LOGGER.debug("Triggered method async_api_load: %s", endpoint)

        # Load the endpoint
        try:
            status, _, content = await self.async_api_query(endpoint, request)
        except AsusRouter404Error:
            _LOGGER.debug("Endpoint %s not found", endpoint)
            return {}
        except AsusRouterAccessError as ex:
            # Check whether we are not connected
            args = ex.args
            if args[1] == AccessError.AUTHORIZATION:
                # Mark the connection as dropped
                await self._async_drop_connection()
                # Wait before repeating the request
                await asyncio.sleep(1 + retry * 3)
                # Repeat request once more and see what happens
                return await self.async_api_load(endpoint, request, True)
            # Otherwise just raise the exception
            raise ex

        # Log status
        _LOGGER.debug("Response %s received from %s", status, endpoint)

        # Try to read the content
        try:
            result = read(endpoint, content)
        except json.JSONDecodeError as ex:
            # Not like this is supposed to happen, but just in case
            _LOGGER.debug("Failed to read content from %s", endpoint)
            _LOGGER.debug("Content: %s", content)
            # Just repeat request once more and see what happens
            # Only if we haven't tried already
            if not retry:
                return await self.async_api_load(endpoint, request, True)
            raise AsusRouterDataError(
                "Something went wrong while reading the content"
            ) from ex

        # Check if we need to drop the connection
        run_service = result.get("run_service", None)
        if run_service in ("restart_httpd", "reboot"):
            await self._async_drop_connection()

        return result

    async def async_api_hook(self, request: str) -> dict[str, Any]:
        """Hook to the device API. Hooks are used to fetch data from the device."""

        _LOGGER.debug("Triggered method async_api_hook: %s", request)

        return await self.async_api_load(
            endpoint=Endpoint.HOOK,
            request=f"hook={request}",
        )

    async def async_api_command(
        self,
        commands: Optional[dict[str, str]],
        endpoint: EndpointType = EndpointControl.COMMAND,
    ):
        """Send a command to the device."""

        _LOGGER.debug(
            "Triggered method async_api_command: %s | %s", endpoint, commands
        )

        return await self.async_api_load(
            endpoint=endpoint,
            request=str(commands),
        )

    # ---------------------------
    # <-- Request-related methods
    # ---------------------------

    def _where_to_get_data(
        self, datatype: AsusData
    ) -> Optional[AsusDataFinder]:
        """Get the list of endpoints to get data from."""

        _LOGGER.debug("Triggered method _where_to_get_data")

        # Check that device identity is available
        if not self._identity:
            _LOGGER.debug("No device identity available")
            return None

        # Get the map
        data_map = ASUSDATA_MAP.get(datatype)
        # Consider aliases
        while isinstance(data_map, AsusData):
            data_map = ASUSDATA_MAP.get(data_map)
        # Check if we have a map
        if not isinstance(data_map, AsusDataFinder):
            _LOGGER.debug("No map found for %s", datatype)
            return None

        # Check if endpoints are available
        for endpoint in data_map.endpoint:
            # Check endpoint availability in identity
            if self._identity.endpoints and self._identity.endpoints.get(
                endpoint
            ) in (
                False,
                None,
            ):
                # Remove the endpoint from the map
                data_map.endpoint.remove(endpoint)

        _LOGGER.debug("Endpoints to check: %s", data_map.endpoint)

        return data_map

    def _transform_data(
        self, datatype: AsusData, data: Any, **kwargs: Any
    ) -> Any:
        """Transform data if needed."""

        _LOGGER.debug("Triggered method _transform_data for `%s`", datatype)

        if datatype == AsusData.CLIENTS:
            _LOGGER.debug("Transforming clients data")
            return transform_clients(
                data,
                self._state.get(AsusData.CLIENTS),
                aimesh=self._identity.aimesh if self._identity else False,
            )

        if datatype == AsusData.CPU:
            _LOGGER.debug("Transforming CPU data")
            return transform_cpu(data)

        if datatype == AsusData.NETWORK:
            _LOGGER.debug("Transforming network data")
            return transform_network(
                data,
                self._identity.services if self._identity else [],
                self._state.get(AsusData.NETWORK),
                model=self._identity.model if self._identity else None,
            )

        if datatype == AsusData.PORTS:
            _LOGGER.debug("Transforming port data")
            return transform_ethernet_ports(
                data,
                self._identity.mac if self._identity else None,
            )

        if datatype == AsusData.WAN:
            _LOGGER.debug("Transforming WAN data")
            return transform_wan(
                data,
                self._identity.services if self._identity else [],
            )

        return data

    def _drop_data(self, datatype: AsusData, endpoint: EndpointType) -> bool:
        """Check whether data should be dropped. This is required
        for some data obtained from multiple endpoints."""

        if not self._identity:
            return False

        if datatype == AsusData.OPENVPN_CLIENT:
            if self._identity.merlin is True:
                return endpoint == Endpoint.HOOK

        return False

    async def _check_prerequisites(self, datatype: AsusData) -> None:
        """Check prerequisites before fetching data."""

        _LOGGER.debug(
            "Triggered method _check_prerequisites for datatype `%s`", datatype
        )

        # A placeholder for future checks
        return

    async def _check_postrequisites(self, datatype: AsusData) -> None:
        """Check postrequisites after fetching data.

        This method is also used to fetch additional data."""

        _LOGGER.debug(
            "Triggered method _check_postrequisites for datatype `%s`",
            datatype,
        )

        # Firmware
        if datatype == AsusData.FIRMWARE:
            # Check if update is available
            firmware = self._state[AsusData.FIRMWARE].data
            if firmware and firmware["state"] is True:
                # Get release notes
                release_note = await self.async_get_data(
                    AsusData.FIRMWARE_NOTE, force=True
                )
                if release_note:
                    firmware.update(release_note)
        return

    def _check_state(self, datatype: Optional[AsusData]) -> None:
        """Make sure the state object is available."""

        _LOGGER.debug("Triggered method _check_state")

        if not datatype:
            return

        # Add state object but make sure it's marked expired
        if datatype not in self._state:
            self._state[datatype] = AsusDataState(
                timestamp=(
                    datetime.now(timezone.utc)
                    - timedelta(seconds=2 * self._cache_time)
                )
            )

    def _return_state(self, datatype: AsusData, **kwargs: Any) -> Any:
        """Return a proper state."""

        _LOGGER.debug("Triggered method _return_state")

        # Get the state
        state = self._state[datatype].data

        if datatype == AsusData.PORTS:
            own_mac = self._identity.mac if self._identity else None

            # Get the device selected
            device = kwargs.get("device", None)

            match device:
                case None:
                    if isinstance(state, dict):
                        return state.get(own_mac, {})
                    return state
                case "all":
                    return state
                # Case when substate is a MAC address
                case a if isinstance(a, str):
                    if isinstance(state, dict) and a in state:
                        return state[a]
                    return {}

        return state

    async def async_get_data(
        self, datatype: AsusData, force: bool = False, **kwargs: Any
    ) -> Any:
        """Generic method to get data from the device."""

        # Check if we have a state object for this data
        self._check_state(datatype)

        # If state object is active, wait for it to finish and return the data
        if self._state[datatype].active:
            try:
                _LOGGER.debug(
                    "Already in progress. Waiting for data to be fetched"
                )
                await asyncio.wait_for(
                    self._state[datatype].inactive_event.wait(),
                    DEFAULT_TIMEOUT,
                )
            except asyncio.TimeoutError:
                _LOGGER.debug(
                    "Timeout while waiting for data. Will try fetching again"
                )

        # Check if we have the data already and not forcing a refresh
        if self._state[datatype].data and not force:
            # Check if the data is younger than the cache time
            if datetime.now(timezone.utc) - self._state[
                datatype
            ].timestamp < timedelta(seconds=self._cache_time):
                _LOGGER.debug(
                    "Using cached data for `%s`: %s",
                    datatype,
                    self._state[datatype].data,
                )
                # Return the cached data
                return self._state[datatype].data
            _LOGGER.debug("Data for %s is too old. Fetching", datatype)

        # Mark the data as active
        self._state[datatype].start()

        # Check prerequisites
        await self._check_prerequisites(datatype)

        # Get the data finder
        data_finder = self._where_to_get_data(datatype)

        # Check if we have a data finder
        if not data_finder:
            _LOGGER.debug("No data finder for %s", datatype)
            return {}

        # The data we are looking for
        data = {}
        result: dict[AsusData, Any] = {}

        try:
            for endpoint in data_finder.endpoint:
                # Get the data from the endpoint
                request = "hook=" if endpoint == Endpoint.HOOK else ""
                for item in data_finder.request:
                    key, value = item
                    request += f"{key}({value});"
                if data_finder.method:
                    argument = self._get_attribute(data_finder.arguments)
                    request += (
                        data_finder.method(argument)
                        if argument
                        else data_finder.method()
                    )
                # Check that we are not fetching this data already

                # Add the request from kwargs
                kw_request = kwargs.get("request", {})
                if isinstance(kw_request, dict):
                    for key, value in kw_request.items():
                        request += f"{key}={value};"
                    # Remove trailing symbol
                    request = request[:-1]

                # Fetch the data
                data = await self.async_api_load(endpoint, request)

                # Make sure, identity is available
                if not self._identity:
                    self._identity = await self.async_get_identity()

                processed = process(
                    endpoint,
                    data,
                    self._state,
                    self._identity.firmware,
                    self._identity.wlan,
                )

                # Check whether data should be dropped
                to_drop = []
                for key, value in processed.items():
                    if self._drop_data(key, endpoint):
                        to_drop.append(key)
                for key in to_drop:
                    processed.pop(key, None)

                result = merge_dicts(result, processed)

                # Check if we have data and data finder merge is ANY
                if result and data_finder.merge == AsusDataMerge.ANY:
                    break

            # Save the data state
            for key, value in result.items():
                # Transform data if needed
                value = self._transform_data(key, value)
                # Save the data
                result[key] = value
                # Update the state
                if key not in self._state:
                    self._state[key] = AsusDataState()
                self._state[key].update(value)
        except (AsusRouterConnectionError, AsusRouterDataError):
            return self._return_state(datatype, **kwargs)

        # Check flags
        await self._check_flags()

        # Check postrequisites
        await self._check_postrequisites(datatype)

        # Return the data we were looking for
        _LOGGER.debug(
            "Returning data for `%s` with object type `%s`",
            datatype,
            type(self._state[datatype].data),
        )
        return self._return_state(datatype, **kwargs)

    # ---------------------------
    # Service-related methods -->
    # ---------------------------

    async def async_run_service(
        self,
        service: Optional[str],
        arguments: Optional[dict[str, Any]] = None,
        apply: bool = False,
        expect_modify: bool = True,
        drop_connection: bool = False,
    ) -> bool:
        """Run a service."""

        _LOGGER.debug("Triggered method async_run_service")

        # Run the service
        result, self._needed_time, self._last_id = await async_call_service(
            self.async_api_command,
            service,
            arguments,
            apply,
            expect_modify,
        )

        if drop_connection:
            await self._async_drop_connection()

        return result

    async def _async_check_state_dependency(self, state: AsusState) -> None:
        """Check and queue state dependencies. Required for some states."""

        _LOGGER.debug("Triggered method _async_check_state_dependency")

        dependency = get_datatype(state)

        if dependency == AsusData.VPNC:
            # VPNC state change requires the correct previous state
            await self.async_get_data(AsusData.VPNC, force=True)

        if dependency == AsusData.AURA:
            # Aura state change requires the correct previous state
            await self.async_get_data(AsusData.AURA, force=True)

        return

    async def _async_get_state_callback(
        self, state: AsusState
    ) -> Callable[..., Awaitable]:
        """Get the state callback."""

        _LOGGER.debug("Triggered method _async_get_state_callback")

        datatype = get_datatype(state)
        # If state is one of AsusState.AURA enum
        if datatype == AsusData.AURA:
            return self.async_api_command

        return self.async_run_service

    async def async_set_state(
        self,
        state: AsusState,
        expect_modify: bool = False,
        **kwargs: Any,
    ) -> bool:
        """Set the state."""

        _LOGGER.debug("Triggered method async_set_state")

        _LOGGER.debug(
            "Setting state `%s` with arguments `%s`. Expecting modify: `%s`",
            state,
            kwargs,
            expect_modify,
        )

        # Check dependencies
        await self._async_check_state_dependency(state)

        # Get the state callback
        callback = await self._async_get_state_callback(state)

        result = await set_state(
            callback=callback,
            state=state,
            expect_modify=expect_modify,
            router_state=self._state,
            identity=self._identity,
            **kwargs,
        )

        # Rewrite the result if it is the default one
        if result == DEFAULT_RESULT_SUCCESS:
            result = True

        if result is True:
            _datatype = get_datatype(state)

            if _datatype in (AsusData.VPNC, AsusData.AURA):
                # The only way to make it work with VPN Fusion
                await asyncio.sleep(1)
                await self._async_check_state_dependency(state)
            elif (
                get_enum_key_by_value(
                    AsusState, type(state), default=AsusState.NONE
                )
                == AsusState.PC_RULE
            ):
                # We should not save this state, since it is saved differently
                pass
            else:
                # Check if we have a state object for this data
                self._check_state(_datatype)
                # Save the state
                _LOGGER.debug(
                    "Saving state `%s` for `%s` s with id=`%s`",
                    state,
                    self._needed_time,
                    self._last_id,
                )
                save_state(
                    state, self._state, self._needed_time, self._last_id
                )
                # Reset the needed time and last id
                self._needed_time = None
                self._last_id = None

        return result

    # ---------------------------
    # <-- Service-related methods
    # ---------------------------

    # ---------------------------
    # Legacy methods -->
    # ---------------------------
    # Not tested, not used, not documented
    # This part can be changed / removed in the future
    # It might also not be working properly

    async def async_apply_port_forwarding_rules(
        self,
        rules: list[PortForwardingRule],
    ) -> bool:
        """Apply port forwarding rules"""

        request = legacy.compile_port_forwarding(rules)

        if request:
            return await self.async_run_service(
                service="restart_firewall",
                arguments=request,
                apply=True,
            )

        return False

    async def async_remove_port_forwarding_rules(
        self,
        ips: str | list[str] | None = None,
        rules: PortForwardingRule | list[PortForwardingRule] | None = None,
        apply: bool = True,
    ) -> list[PortForwardingRule]:
        """Remove port forwarding rules"""

        ips = set() if ips is None else set(safe_list(ips))
        rules = [] if rules is None else safe_list(rules)

        # Get current rules
        current_rules: list[PortForwardingRule] = (
            await self.async_get_data(AsusData.PORT_FORWARDING)
        )["rules"]

        # Remove all rules for these IPs
        current_rules = [
            rule for rule in current_rules if rule.ip_address not in ips
        ]
        # Remove exact rules
        for rule_to_find in rules:
            current_rules = [
                rule
                for rule in current_rules
                if not (
                    rule.ip_address == rule_to_find.ip_address
                    and rule.port_external == rule_to_find.port_external
                    and rule.protocol == rule_to_find.protocol
                    and (
                        rule_to_find.ip_external is None
                        or rule.ip_external == rule_to_find.ip_external
                    )
                    and (
                        rule_to_find.port is None
                        or rule.port == rule_to_find.port
                    )
                )
            ]

        # Apply new rules
        if apply:
            await self.async_apply_port_forwarding_rules(current_rules)

        # Return the new rules
        return current_rules

    async def async_set_port_forwarding_rules(
        self,
        rules: PortForwardingRule | list[PortForwardingRule],
    ) -> bool:
        """Set port forwarding rules"""

        rules = safe_list(rules)

        # Get current rules
        current_rules: list[PortForwardingRule] = (
            await self.async_get_data(AsusData.PORT_FORWARDING)
        )["rules"]

        # Add new rules
        for rule in rules:
            current_rules.append(rule)

        # Apply new rules
        return await self.async_apply_port_forwarding_rules(current_rules)

    # ---------------------------
    # <-- Legacy methods
    # ---------------------------

    # ---------------------------
    # Properties -->
    # ---------------------------

    @property
    def connected(self) -> bool:
        """Return connection status."""

        return self._connection.connected if self._connection else False

    # ---------------------------
    # <-- Properties
    # ---------------------------

    # ---------------------------
    # Additional settings -->
    # ---------------------------

    # ---------------------------
    # <-- Additional settings
    # ---------------------------

    # ---------------------------
    # General management -->
    # ---------------------------

    async def async_cleanup(self) -> None:
        """Cleanup the connection."""

        if self._connection:
            self._connection.reset_connection()
            # await self._connection._async_close_session()

    # ---------------------------
    # <-- General management
    # ---------------------------



"""WLAN module."""

from __future__ import annotations

from enum import Enum, IntEnum
from typing import Any, Awaitable, Callable

from asusrouter.modules.const import MapValueType
from asusrouter.tools.converters import (
    get_arguments,
    safe_bool,
    safe_int,
    safe_unpack_key,
)
from asusrouter.tools.writers import nvram


class Wlan(str, Enum):
    """WLAN type class."""

    FREQ_2G = "2ghz"
    FREQ_5G = "5ghz"
    FREQ_5G2 = "5ghz2"
    FREQ_6G = "6ghz"
    UNKNOWN = "unknown"


# A map to correspond possible values to the WlanType
# E.g. `wlc_0` -> FREQ_2G, `wlc_1` -> FREQ_5G, etc.
# But also `wl0` -> FREQ_2G, `wl1` -> FREQ_5G, etc.
WLAN_TYPE: dict[str | int, Wlan] = {
    0: Wlan.FREQ_2G,
    1: Wlan.FREQ_5G,
    2: Wlan.FREQ_5G2,
    3: Wlan.FREQ_6G,
    "2.4G": Wlan.FREQ_2G,
    "5G": Wlan.FREQ_5G,
    "5G-2": Wlan.FREQ_5G2,
    "wifi6e": Wlan.FREQ_6G,
    "wlc_0": Wlan.FREQ_2G,
    "wlc_1": Wlan.FREQ_5G,
    "wlc_2": Wlan.FREQ_5G2,
    "wlc_3": Wlan.FREQ_6G,
    "wl0": Wlan.FREQ_2G,
    "wl1": Wlan.FREQ_5G,
    "wl2": Wlan.FREQ_5G2,
    "wl3": Wlan.FREQ_6G,
}

# A map of NVRAM values for a GWLAN
MAP_GWLAN: list[MapValueType] = [
    ("wl{}_akm"),
    ("wl{}_ap_isolate"),
    ("wl{}_auth"),
    ("wl{}_auth_mode"),
    ("wl{}_auth_mode_x"),
    ("wl{}_bridge"),
    ("wl{}_bss_enabled", safe_bool),
    ("wl{}_bss_maxassoc", safe_int),
    ("wl{}_bw_dl", safe_int),  # Bandwidth limit download
    ("wl{}_bw_enabled", safe_bool),  # Bandwidth limit switch
    ("wl{}_bw_ul", safe_int),  # Bandwidth limit upload
    ("wl{}_closed", safe_bool),
    ("wl{}_crypto"),
    ("wl{}_expire", safe_int),  # Expire time in s
    ("wl{}_expire_tmp", safe_int),  # Expire time left in s
    ("wl{}_gn_wbl_enable"),
    ("wl{}_gn_wbl_rule"),
    ("wl{}_gn_wbl_type"),
    ("wl{}_hwaddr"),  # MAC address
    ("wl{}_ifname"),  # Interface name
    ("wl{}_infra"),
    ("wl{}_key"),
    ("wl{}_key1"),
    ("wl{}_key2"),
    ("wl{}_key3"),
    ("wl{}_key4"),
    ("wl{}_lanaccess", safe_bool),  # LAN access
    ("wl{}_maclist"),
    ("wl{}_macmode"),
    ("wl{}_maxassoc", safe_int),
    ("wl{}_mbss"),
    ("wl{}_mfp"),
    ("wl{}_mode"),
    ("wl{}_net_reauth", safe_int),
    ("wl{}_preauth"),
    ("wl{}_radio", safe_bool),
    ("wl{}_radius_ipaddr"),
    ("wl{}_radius_key"),
    ("wl{}_radius_port", safe_int),
    ("wl{}_sae_anti_clog_threshold"),
    ("wl{}_sae_groups"),
    ("wl{}_sae_sync"),
    ("wl{}_ssid"),  # SSID
    ("wl{}_sta_retry_time"),
    ("wl{}_sync_node", safe_bool),  # Sync AiMesh nodes
    ("wl{}_unit"),  # GWLAN unit id
    ("wl{}_wep", safe_bool),
    ("wl{}_wep_x", safe_bool),
    ("wl{}_wfi_enable", safe_bool),
    ("wl{}_wfi_pinmode"),
    ("wl{}_wme"),
    ("wl{}_wme_bss_disable", safe_bool),
    ("wl{}_wpa_gtk_rekey"),
    ("wl{}_wpa_psk"),  # Password
    ("wl{}_wps_mode"),
]


# A map of NVRAM values for a WLAN
MAP_WLAN: list[MapValueType] = [
    ("wl{}_auth_mode_x"),
    ("wl{}_bw"),
    ("wl{}_channel"),
    ("wl{}_chanspec"),
    ("wl{}_closed", safe_bool),
    ("wl{}_country_code"),
    ("wl{}_crypto"),
    ("wl{}_gmode_check"),
    ("wl{}_maclist_x"),
    ("wl{}_macmode"),
    ("wl{}_mbo_enable"),
    ("wl{}_mfp"),
    ("wl{}_nmode_x"),
    ("wl{}_optimizexbox_ckb"),
    ("wl{}_radio", safe_bool),
    ("wl{}_radius_ipaddr"),
    ("wl{}_radius_key"),
    ("wl{}_radius_port", safe_int),
    ("wl{}_ssid"),
    ("wl{}_wpa_gtk_rekey"),
    ("wl{}_wpa_psk"),
]


class AsusWLAN(IntEnum):
    """Asus WLAN state."""

    UNKNOWN = -999
    OFF = 0
    ON = 1


def _nvram_request(
    wlan: list[Wlan] | None, mapping: list[MapValueType], guest: bool = False
) -> str | None:
    """Create an NVRAM request."""

    if not wlan:
        return None

    request = []

    for interface in wlan:
        for pair in mapping:
            key, _ = safe_unpack_key(pair)
            index = list(Wlan).index(interface)
            if guest:
                for gid in range(1, 4):
                    request.append(key.format(f"{index}.{gid}"))
            else:
                request.append(key.format(index))

    return nvram(request)


def wlan_nvram_request(wlan: list[Wlan] | None) -> str | None:
    """Create an NVRAM request for WLAN."""

    return _nvram_request(wlan, MAP_WLAN)


def gwlan_nvram_request(wlan: list[Wlan] | None) -> str | None:
    """Create an NVRAM request for GWLAN."""

    return _nvram_request(wlan, MAP_GWLAN, guest=True)


async def set_state(
    callback: Callable[..., Awaitable[bool]],
    state: AsusWLAN,
    **kwargs: Any,
) -> bool:
    """Set the parental control state."""

    # Get the arguments
    api_type, api_id = get_arguments(("api_type", "api_id"), **kwargs)

    # Check if the api_type and api_id are available
    if api_type is None or api_id is None:
        return False

    # Define the service and callback arguments for each api_type
    api_values = {
        "wlan": {
            "service": "restart_wireless",
            "callback_arguments": {
                f"wl{api_id}_radio": 1 if state == AsusWLAN.ON else 0
            },
        },
        "gwlan": {
            "service": "restart_wireless;restart_firewall",
            "callback_arguments": {
                f"wl{api_id}_bss_enabled": 1 if state == AsusWLAN.ON else 0,
                **({f"wl{api_id}_expire": 0} if state == AsusWLAN.ON else {}),
            },
        },
    }

    # Get the service and callback arguments for the given api_type
    api_value = api_values.get(api_type)
    if not api_value:
        return False

    # Call the service
    return await callback(
        api_value["service"],
        arguments=api_value["callback_arguments"],
        apply=True,
        expect_modify=kwargs.get("expect_modify", False),
    )



"""Tests for the configuration module."""

from __future__ import annotations

import pytest
from asusrouter.config import CONFIG_DEFAULT_BOOL, ARConfig, safe_bool_config

KEYS_BOOL = [
    "optimistic_data",
]


@pytest.fixture(autouse=True)
def reset_config() -> None:
    """Reset the configuration before each test."""

    ARConfig.set("optimistic_data", CONFIG_DEFAULT_BOOL)
    ARConfig.set("optimistic_temperature", CONFIG_DEFAULT_BOOL)


class TestConvert:
    """Tests for the configuration conversion functions."""

    @pytest.mark.parametrize("value", [True, False])
    def test_safe_bool_config(self, value: bool) -> None:
        """Test that safe_bool_config converts values correctly."""

        assert safe_bool_config(value) is value

    def test_safe_bool_config_default(self) -> None:
        """Test that safe_bool_config returns default when value is None."""

        assert safe_bool_config(None) is CONFIG_DEFAULT_BOOL


class TestBoolConfig:
    """Tests for boolean configuration options."""

    def test_keys(self) -> None:
        """Test that we can get the full list of configuration keys."""

        keys = ARConfig.keys()
        assert isinstance(keys, list)
        assert len(keys) > 0
        assert all(isinstance(key, str) for key in keys)

    @pytest.mark.parametrize("key", KEYS_BOOL)
    def test_default_bool(self, key: str) -> None:
        """Test the default value of a boolean configuration key."""

        assert ARConfig.get(key) is CONFIG_DEFAULT_BOOL

    @pytest.mark.parametrize("key", KEYS_BOOL)
    @pytest.mark.parametrize("value", [True, False])
    def test_set_bool(self, key: str, value: bool) -> None:
        """Test setting a boolean configuration key."""

        ARConfig.set(key, value)
        assert ARConfig.get(key) is value

    def test_get_optimistic_data(self) -> None:
        """Test that we can get the value of optimistic_data."""

        assert ARConfig.optimistic_data is CONFIG_DEFAULT_BOOL

    def test_get_optimistic_temperature(self) -> None:
        """Test that we can get the value of optimistic_temperature."""

        assert ARConfig.optimistic_temperature is CONFIG_DEFAULT_BOOL

    def test_unknown_key(self) -> None:
        """Test that setting/getting an unknown key raises KeyError."""

        with pytest.raises(KeyError):
            ARConfig.set("unknown_option", True)

        with pytest.raises(KeyError):
            ARConfig.get("unknown_option")



"""WireGuard module."""

from __future__ import annotations

import logging
from enum import IntEnum
from typing import Any, Awaitable, Callable, Optional

_LOGGER = logging.getLogger(__name__)


class AsusWireGuardClient(IntEnum):
    """Asus WireGuard client state."""

    UNKNOWN = -999
    OFF = 0
    ON = 1


class AsusWireGuardServer(IntEnum):
    """Asus WireGuard server state."""

    UNKNOWN = -999
    OFF = 0
    ON = 1


def _get_arguments(**kwargs: Any) -> Optional[int]:
    """Get the arguments from kwargs."""

    arguments = kwargs.get("arguments", {})

    # Get the id from arguments
    wlan_id = arguments.get("id") if arguments else kwargs.get("id")
    if wlan_id is None:
        wlan_id = 1
        _LOGGER.debug("Using default id 1")

    return wlan_id


async def set_state(
    callback: Callable[..., Awaitable[bool]],
    state: AsusWireGuardClient | AsusWireGuardServer,
    **kwargs: Any,
) -> bool:
    """Set the WireGuard state."""

    # Get the arguments
    wlan_id = _get_arguments(**kwargs)

    # WireGuard unit type (server or client)
    wg_unit = "wgs" if isinstance(state, AsusWireGuardServer) else "wgc"

    # Callback arguments
    callback_arguments = {
        "id": wlan_id,
        f"{wg_unit}_enable": 1
        if state in (AsusWireGuardClient.ON, AsusWireGuardServer.ON)
        else 0,
        f"{wg_unit}_unit": wlan_id,
    }

    # Get the expect_modify argument
    expect_modify = kwargs.get("expect_modify", False)

    # Get the correct service call
    service_map: dict[Any, str] = {
        (AsusWireGuardClient, AsusWireGuardClient.ON): f"start_wgc {wlan_id}",
        (AsusWireGuardClient, AsusWireGuardClient.OFF): f"stop_wgc {wlan_id}",
        (AsusWireGuardServer, AsusWireGuardServer.ON): "restart_wgs;restart_dnsmasq",
        (AsusWireGuardServer, AsusWireGuardServer.OFF): "restart_wgs;restart_dnsmasq",
    }

    service = service_map.get((type(state), state))

    if not service:
        _LOGGER.debug("Unknown state %s", state)
        return False

    # Call the service
    return await callback(
        service=service,
        arguments=callback_arguments,
        apply=True,
        expect_modify=expect_modify,
    )
